#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/mappings.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.06.2018
# Last Modified Date: 17.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import json
from typing import Dict, Any, List, Union, Type, Optional, Sequence, TypeVar, Literal, Callable, Iterable, Mapping, MutableMapping
from ampel.type import strict_iterable
from ampel.util.crypto import hash_payload, HT
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.util.collections import check_seq_inner_type
from ampel.util.crypto import b2_short_hash

T = TypeVar('T')

def build_unsafe_short_dict_id(dict_arg: Optional[Dict]) -> int:
	"""
	Note: no collision occured applying blake2 using 7bytes digests on the word list
	https://github.com/dwyl/english-words/blob/master/words.txt
	containing 466544 english words
	:returns: 7bytes int (MongoDB supports only *signed* 64-bit integers)

	example:
	In []: build_unsafe_short_dict_id({'a': 1})
	Out[]: 18043533495046284

	In []: build_unsafe_short_dict_id({'b': 2, 'a': 1, 'c': {'b': 3, 'a': [4, 'r', 1]}})
	Out[]: 53310293724158701

	In []: build_unsafe_short_dict_id({'a': 1, 'b': 2, 'c': {'a': [4, 1, 'r'], 'b': 3}})
	Out[]: 53310293724158701
	"""
	return build_unsafe_dict_id(dict_arg, int, 'blake2b', digest_size=7)


def build_unsafe_dict_id(
	dict_arg: Optional[Dict],
	ret: Type[HT] = bytes, # type: ignore[assignment]
	alg: Literal['sha512', 'sha1', 'blake2b'] = 'sha512',
	sort_keys: bool = True,
	flatten_list_members: bool = True,
	sort_lists: bool = True,
	flatten_lists: bool = True,
	**kwargs
) -> HT:
	"""
	:param dict_arg: can be nested, can be None
	:param ret: return type, can be bytes, str (hex digest) or int
	:param alg: hash algorithm (default is sha512)
	:param sort_keys: see `flatten_dict` docstring
	:param flatten_list_members: see `flatten_dict` docstring
	:param sort_lists: see `flatten_dict` docstring
	:param flatten_lists: see `flatten_dict` docstring
	:param kwargs: will be forwarded to hashlib hash function

	example:
	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'a': ['r', 1, 4], 'b': 3}}, ret=str)
	Out[]: 'b5acfa0d427fe1ef682895217c94400178b5700997a9547fe5bebf33b73d8157332c2bb1bd0e370
	0c8c232fd55f4993b1b34132420afc14a05e2414df3037519'

	In []: build_unsafe_dict_id({'b': 2, 'a': 1, 'c': {'b': 3, 'a': [4, 'r', 1]}}, ret=str)
	Out[]: 'b5acfa0d427fe1ef682895217c94400178b5700997a9547fe5bebf33b73d8157332c2bb1bd0e370
	0c8c232fd55f4993b1b34132420afc14a05e2414df3037519'

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 5]}}, ret=str)
	Out[]: '1f8e8c35e9641a6f8cb5a1136b712cf1b735577645db6e1ee373ea7dd08266b63a8b23fde0
	615c6916205cfdf928e42cd79171581c211eb77bed967d65563b2f'

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 4]}}, ret=int, alg='sha1')
	Out[]: 967659017817567346241766354309619194352380159869

	In []: build_unsafe_dict_id({'a': 1, 'b': 2}, ret=int, alg='blake2b', digest_size=7)
	Out[]: 32414584742937293
	"""

	if dict_arg is None:
		dict_arg = {}

	return hash_payload(
		bytes(
			json.dumps(
				flatten_dict(
					dict_arg, '.', sort_keys, flatten_list_members,
					sort_lists, flatten_lists
				),
				indent=None, separators=(',', ':')
			),
			"utf8"
		),
		ret, alg, **kwargs
	)


def get_by_path(
	mapping: Dict, path: Union[str, Sequence[str]], delimiter: str = '.'
) -> Optional[Any]:
	"""
	Get an item from a nested mapping by path, e.g.
	'foo.bar.baz' -> mapping['foo']['bar']['baz']

	:param path: example: 'foo.bar.baz' or ['foo', 'bar', 'baz']
	:param delimiter: example: '.'
	"""

	if isinstance(path, str):
		path = path.split(delimiter)

	# check for int elements encoded as str
	path: List[Union[int, str]] = [ # type: ignore
		(el if not el.isdigit() else int(el)) for el in path
	]

	for el in path:
		if el not in mapping:
			return None
		mapping = mapping[el]

	return mapping


def set_by_path(
	d: Dict, path: Union[str, Sequence[str]], val: Any,
	delimiter: str = '.', create: bool = True
) -> bool:
	"""
	:param create: whether to create directory sub-structures if they do not exits
	(in this case, this method will alawys return False)
	:returns: False if the key was successfully set, True otherwise
	"""
	if isinstance(path, str):
		path = path.split(delimiter) # type: ignore
	l = len(path) - 1
	for i, k in enumerate(path):
		if k not in d:
			if not create:
				return True
			d[k] = {}
		if i == l:
			d[k] = val
			return False
		d = d[k]
	return True


def del_by_path(d: Dict, path: Union[str, Sequence[str]], delimiter: str = '.') -> bool:
	""" :returns: False if the key was successfully deleted, True otherwise """

	if isinstance(path, str):
		path = path.split(delimiter) # type: ignore
	l = len(path) - 1
	for i, k in enumerate(path):
		if k not in d:
			return True
		if i == l:
			del d[k]
			return False
		d = d[k]
	return True


def walk_and_process_dict(
	arg: Union[dict, list], callback: Callable,
	match: List[str], path: str = None, **kwargs
) -> bool:
	"""
	callback is called with 4 arguments:
	1) the path of the possibly nested entry. Ex: 'processor.config.select' or 'processor'
	2) the matching key (from list 'match'). Ex: 'config'
	3) the matching (sub) dict
	4) the **kwargs provided to this method
	and should return True if a modification was performed

	Simplest callback function:
	def my_callback(path, k, d):
		print(f'{path} -> {k}: {d}\n')
		return False

	:returns: True if a modification was performed, False otherwise
	"""

	ret = False

	if isinstance(arg, list):
		for i, el in enumerate(arg):
			ret = walk_and_process_dict(
				el, callback, match, f'{path}.{i}' if path else f'{i}', **kwargs
			) or ret

	if isinstance(arg, dict):

		for k, v in arg.items():

			if k in match:
				ret = callback(path, k, arg, **kwargs) or ret

			if isinstance(v, dict):
				ret = walk_and_process_dict(
					v, callback, match, f'{path}.{k}' if path else f'{k}', **kwargs
				) or ret

			if isinstance(v, list):
				for i, el in enumerate(v):
					ret = walk_and_process_dict(
						el, callback, match, f'{path}.{k}.{i}' if path else f'{k}.{i}', **kwargs
					) or ret

	return ret

def flatten_dict(
	d: Mapping,
	separator: str = '.',
	sort_keys: bool = False,
	flatten_list_members: bool = False,
	flatten_lists: bool = False,
	sort_lists: bool = False
) -> MutableMapping:
	"""
	This function is useful, among other things, for building "hash ids" of serializable dicts

	:param separator: character to be used to concatenate dict keys of different levels: {'a': {'b': 1}} -> {'a.b': 1}
	:param sort_keys: whether to sort dict keys. This applies to all dicts regardless of their depth/nesting
	:param flatten_list_members: whether to flatten dict structures embedded in list/sequences
	:param flatten_lists: whether to flatten lists, effectively converting ['a', 'b'] into {'0': 'a', '2': 'b'}
	:param sort_lists: whether to sort lists when possible, effectively converting ['r', 'a', 4, 1] into [1, 4, 'a', 'r']

	Example:
	Simplest case:
	In []: flatten_dict({'count': {'chans': {'HU_SN': 10}}})
	Out[]: {'count.chans.HU_SN': 10}

	In []: flatten_dict({'d': {'e':1}, 'a': [{'c':2}, {'b':{'f':[3, 1, 2]}}]}, sort_keys=True)
	Out[]: {'a': [{'c': 2}, {'b': {'f': [3, 1, 2]}}], 'd.e': 1}

	In []: flatten_dict({'d': {'e':1}, 'a': [{'c':2}, {'b':{'f':[3, 1, 2]}}]}, sort_keys=True, flatten_list_members=True)
	Out[]: {'a': [{'c': 2}, {'b.f': [3, 1, 2]}], 'd.e': 1}

	In []: flatten_dict({'d': {'e':1}, 'a': [{'c':2}, {'b':{'f':[3, 1, 2]}}]}, sort_keys=True, flatten_list_members=True, sort_lists=True)
	Out[]: {'a': [{'b.f': [1, 2, 3]}, {'c': 2}], 'd.e': 1}

	In []: flatten_dict({'d': {'e':1}, 'a': [{'b':{'f': [1, 2, 3]}}, {'c':2}]}, sort_keys=True, flatten_list_members=True, sort_lists=True)
	Out[]: {'a': [{'b.f': [1, 2, 3]}, {'c': 2}], 'd.e': 1}

	In []: flatten_dict({'d': {'e':1}, 'a': [{'b':{'f': [1, 2, 3]}}, {'c':2}]}, sort_keys=True, flatten_list_members=True, sort_lists=True, flatten_lists=True)
	Out[]: {'a.0.b.f.0': 1, 'a.0.b.f.1': 2, 'a.0.b.f.2': 3, 'a.1.c': 2, 'd.e': 1}

	In []: flatten_dict({'d': {'e':1}, 'a': [{'c':2}, {'b':{'f':[3, 1, 2]}}]}, sort_keys=True, flatten_list_members=True, sort_lists=True, flatten_lists=True)
	Out[]: {'a.0.b.f.0': 1, 'a.0.b.f.1': 2, 'a.0.b.f.2': 3, 'a.1.c': 2, 'd.e': 1}
	"""

	try:
		out = {}
		for k in sorted(d.keys()) if sort_keys else d:

			v = d[k]

			if isinstance(v, dict):

				for kk, vv in flatten_dict(
					v, separator, sort_keys, flatten_list_members, flatten_lists, sort_lists
				).items():
					out[f'{k}{separator}{kk}'] = vv

			elif isinstance(v, strict_iterable):

				if flatten_list_members:
					v = [
						flatten_dict(el, separator, sort_keys, flatten_list_members, flatten_lists, sort_lists)
						if isinstance(el, dict) else el
						for el in v
					]

				if sort_lists:

					try:
						# allow int/str mixed up
						v = sorted(v, key=lambda x: str(x))
					except Exception:
						pass

					# In []: sorted([{'c': 2}, {'b.f.0': 1, 'b.f.1': 2, 'b.f.2': 3}], key=lambda x: next(iter(x.keys())))
					# Out[]: [{'b.f.0': 1, 'b.f.1': 2, 'b.f.2': 3}, {'c': 2}]
					if flatten_list_members and all(isinstance(el, dict) for el in v):
						v = sorted(v, key=lambda x: next(iter(x.keys())))

				if flatten_lists:
					for kk, vv in flatten_dict(
						{i: v[i] for i in range(len(v))},
						separator, sort_keys, flatten_list_members, flatten_lists, sort_lists
					).items():
						out[f'{k}{separator}{kk}'] = vv

				else:
					out[k] = v
			else:
				out[k] = v

		return out

	except Exception as e:
		raise ValueError(f"Offending input: {d}") from e


def unflatten_dict(d: Mapping[str, Any], separator: str = '.', unflatten_list: bool = False) -> MutableMapping[str, Any]:
	"""
	Example:

	In []: unflatten_dict({'count.chans.HU_SN': 10})
	Out[]: {'count': {'chans': {'HU_SN': 10}}}

	In []: unflatten_dict({'a.0.b.f.0': 1, 'a.0.b.f.1': 2, 'a.0.b.f.2': 3, 'a.1.c': 2, 'd.e': 1}, unflatten_list=True)
	Out[]: {'a': [{'b': {'f': [1, 2, 3]}}, {'c': 2}], 'd': {'e': 1}}
	"""
	out: Dict[str, Any] = {}

	for key, value in d.items():

		parts = key.split(separator)
		target: Dict[str, Any] = out

		for part in parts[:-1]:
			if part not in target:
				target[part] = {}
			target = target[part]

		target[parts[-1]] = value

	if unflatten_list:
		return _unflatten_lists(out)

	return out


def _unflatten_lists(d: Dict) -> Dict:
	"""
	Note: modifies dict

	In []: _unflatten_lists({'a': {'0': {'b': {'f': {'0': 1, '1': 2, '2': 3}}}, '1': {'c': 2}}, 'd': {'e': 1}})
	Out[]: {'a': [{'b': {'f': [1, 2, 3]}}, {'c': 2}], 'd': {'e': 1}}
	"""

	for k, v in d.items():
		try:
			# Following line's purpose is just to trigger an error when needed:
			# it only works if v is a dict whose keys are integer (all of them)
			[int(kk) for kk in v]
			d[k] = [
				_unflatten_lists(d[k][kk]) if isinstance(d[k][kk], dict) else d[k][kk]
				for kk in v
			]
		except Exception:
			if isinstance(v, dict):
				d[k] = _unflatten_lists(v)

	return d


def merge_dict(d1: Dict, d2: Dict) -> Dict:
	k1 = set(d1.keys())
	k2 = set(d2.keys())
	return {
		**{k: d1[k] for k in k1.difference(k2)},
		**{k: d2[k] for k in k2.difference(k1)},
		**{
			k: merge_dict(d1[k], d2[k]) if isinstance(d1[k], dict) else d2[k]
			for k in k1.intersection(k2)
		}
	}


def merge_dicts(items: Sequence[Optional[Dict[T, Any]]]) -> Optional[Dict[T, Any]]:
	"""
	Merge a sequence of dicts recursively. Elements that are None are skipped.
	"""
	left = None
	for right in items:
		if left and right:
			left = merge_dict(left, right)
		elif right or left is None:
			left = right
	return left


def compare_dict_values(d1: Dict, d2: Dict, keys: Iterable[str]) -> bool:
	"""
	:returns: true if the values of dict one and two are equal for all keys requested
	Note: dict keys absent in both dicts mean that both dicts are equals wrt the dict key.

	In []: compare_dict_values({'a': 1}, {'b': 1}, ['a'])
	Out[56]: False

	In []: compare_dict_values({'a': 1}, {'a': 2}, ['a'])
	Out[]: False

	In []: compare_dict_values({'a': 1}, {'a': 1}, ['a'])
	Out[]: True

	In []: compare_dict_values({'a': 1}, {'a': 1}, ['a', 'b'])
	Out[]: True
	"""

	for f in keys:
		if f in d1:
			if f in d2:
				if d1[f] != d2[f]:
					return False
			else:
				return False
		else:
			if f in d2:
				return False
	return True


def get_nested_attr(obj, path):
	"""
	Get a nested attribute from object:

	:param Object obj:
	:param str path: example: 'foo.bar.baz'
	:rtype: object or None

	.. sourcecode:: python\n
		In []: time_constraint_config.before.value
		Out[]: 1531306299

		In []: AmpelUtils.get_nested_attr(time_constraint_config, "before.value")
		Out[]: 1531306299
	"""
	try:
		for name in path.split("."):
			obj = getattr(obj, name)
		return obj
	except AttributeError:
		return None


def hash_logic_schema(
	arg: Optional[Union[str, dict, AllOf, AnyOf, OneOf]]
) -> Union[int, dict]:
	"""
	Converts dict schema containing str representation of tags into
	a dict schema containing hashed values (int64).

	:param arg: schema dict. \
	See :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` \
	docstring for more details

	examples:
	In []: hash_logic_schema('aa')
	Out[]: 24517795197330556

	In []: hash_logic_schema({'allOf': ['aa', 'bb', 12]})
	Out[]: {'allOf': [24517795197330556, 14271023143587293, 12]}

	In []: hash_logic_schema({'anyOf': ['aa', 'bb', 12]})
	Out[]: {'anyOf': [24517795197330556, 14271023143587293, 12]}

	In []: hash_logic_schema({'anyOf': [{'allOf': ['aa', 'bb', 100]}, 'cc']})
	Out[]: {'anyOf': [{'allOf': [24517795197330556, 14271023143587293, 100]}, 59944183417054336]}

	:returns: new schema dict where tag elements are integers
	"""

	out: Dict[str, Any] = {}

	if isinstance(arg, str):
		return b2_short_hash(arg)

	if isinstance(arg, (AllOf, AnyOf, OneOf)):
		arg = arg.dict()

	if isinstance(arg, dict):

		if 'anyOf' in arg:
			if check_seq_inner_type(arg['anyOf'], str):
				out['anyOf'] = _hash_elements(arg['anyOf'])
			else:
				out['anyOf'] = []
				for el in arg['anyOf']:
					if isinstance(el, str):
						out['anyOf'].append(b2_short_hash(el))
					elif isinstance(el, dict):
						if 'allOf' not in el:
							raise ValueError('Unsupported format (1)')
						out['anyOf'].append(
							{'allOf': _hash_elements(el['allOf'])}
						)
					else:
						out['anyOf'].append(el)

		elif 'allOf' in arg:
			out['allOf'] = _hash_elements(arg['allOf'])

		elif 'oneOf' in arg:
			out['oneOf'] = _hash_elements(arg['oneOf'])
	else:
		raise ValueError(f'Unsupported argument type: "{type(arg)}"')

	return out


def _hash_elements(seq: Sequence) -> List:

	return [
		b2_short_hash(el) if isinstance(el, str) else el
		for el in seq
	]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/select/T3StockSelector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.12.2019
# Last Modified Date: 20.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pymongo.cursor import Cursor
from typing import Union, Optional, Dict, Literal

from ampel.type import ChannelId, Tag
from ampel.db.query.stock import build_stock_query
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.log.utils import safe_query_dict
from ampel.t3.select.AbsT3Selector import AbsT3Selector
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.time.TimeConstraintModel import TimeConstraintModel
from ampel.config.LogicSchemaUtils import LogicSchemaUtils


class T3StockSelector(AbsT3Selector):
	"""
	Default stock selector used by T3Processor
	
	Example configuration::
	  
	  {
	     "created": {"after": {"use": "$timeDelta", "arguments": {"days": -40}}},
	     "modified": {"after": {"use": "$timeDelta", "arguments": {"days": -1}}},
	     "channel": "HU_GP_CLEAN",
	     "tags": {"with": "ZTF", "without": "HAS_ERROR"}
	  }
	"""

	logger: AmpelLogger
	#: Select by creation time
	created: Optional[TimeConstraintModel] = None
	#: Select by modification time
	modified: Optional[TimeConstraintModel] = None
	#: Select by channel
	channel: Optional[Union[ChannelId, AnyOf[ChannelId], AllOf[ChannelId], OneOf[ChannelId]]] = None
	#: Select by tag
	tag: Optional[Dict[Literal['with', 'without'], Union[Tag, Dict, AllOf[Tag], AnyOf[Tag], OneOf[Tag]]]] = None


	def __init__(self, **kwargs):

		if 'channel' in kwargs:
			kwargs['channel'] = LogicSchemaUtils.to_logical_struct(kwargs['channel'], 'channel')

		if 'tag' in kwargs:
			kwargs['tag'] = {
				k: LogicSchemaUtils.to_logical_struct(v, 'tag')
				for k,v in kwargs['tag'].items()
			}

		super().__init__(**kwargs)
		if self.logger is None:
			raise ValueError("Parameter logger cannot be None")


	# Override/Implement
	def fetch(self) -> Optional[Cursor]:

		# Build query for matching transients using criteria defined in config
		match_query = build_stock_query(
			channel = self.channel,
			tag = self.tag,
			time_created = self.created.get_query_model(db=self.context.db) \
				if self.created else None,
			time_modified = self.modified.get_query_model(db=self.context.db) \
				if self.modified else None,
		)

		if self.logger.verbose:
			self.logger.log(VERBOSE, "Executing search query", extra=safe_query_dict(match_query))

		# Execute 'find transients' query
		cursor = self.context.db \
			.get_collection('stock') \
			.find(
				match_query,
				{'_id': 1}, # indexed query
				no_cursor_timeout = True, # allow query to live for > 10 minutes
			) \
			.hint('_id_1_channel_1')

		# Count results
		if cursor.count() == 0:
			self.logger.info("No transient matches the given criteria")
			return None

		self.logger.info(
			f"{cursor.count()} transients match search criteria"
		)

		return cursor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/db/query/utils.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 31.10.2018
# Last Modified Date: 21.11.2019
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any
from ampel.type import strict_iterable, StrictIterable


def add_or(query: Dict[str, Any], arg: Dict[str, Any]) -> None:

	if '$or' in query:
		if '$and' in query:
			if not isinstance(query['$and'], list):
				raise ValueError(f"Illegal $and value in query: {query}")
			query['$and'].append(arg)
		else:
			query['$and'] = [query.pop('$or'), arg]

	else:
		query['$or'] = arg


def match_array(arg: StrictIterable[Any]):
	"""
	match_array(['ab']) -> returns 'ab'
	match_array({'ab'}) -> returns 'ab'
	match_array(['a', 'b']) -> returns {$in: ['a', 'b']}
	match_array({'a', 'b'}) -> returns {$in: ['a', 'b']}
	"""

	if not isinstance(arg, strict_iterable):
		raise ValueError(
			f"Provided argument is not sequence ({type(arg)})"
		)

	if len(arg) == 1:
		return next(iter(arg))

	if isinstance(arg, list):
		return {'$in': arg}

	# Otherwise cast to list
	return {'$in': list(arg)}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/builder/ProcessMorpher.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 01.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import json
from typing import Dict, Any, Literal, Optional
from importlib import import_module
from pydantic import create_model
from pydantic.main import ModelMetaclass
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.model.StrictModel import StrictModel
from ampel.model.ProcessModel import ProcessModel
from ampel.model.Secret import Secret
from ampel.base.DataUnit import DataUnit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.util.mappings import walk_and_process_dict
from ampel.util.type_analysis import get_subtype


class ProcessMorpher:
	""" Applies various transformations to process dicts """

	def __init__(self,
		process: Dict[str, Any],
		templates: Dict[str, Any],
		logger: AmpelLogger,
		verbose: bool = False,
		deep_copy: bool = False
	) -> None:

		self.process = json.loads(json.dumps(process)) if deep_copy else process
		self.templates = templates
		self.logger = logger
		self.verbose = verbose


	def get(self) -> Dict[str, Any]:
		""" :raises: Error if the morphed process does not comply with ProcessModel """
		return ProcessModel(**self.process).dict()


	def enforce_t3_channel_selection(self, chan_name: str) -> 'ProcessMorpher':

		if self.process['tier'] == 3:

			for directive in self.process['processor']['config']['directives']:
				if 'select' in directive:
					select = directive['select']
					if select['unit'] not in ('T3StockSelector', 'T3FilteringStockSelector'):
						raise ValueError(
							f'Cannot enforce channel selection: '
							f'unknown stock selection unit {select["unit"]}'
						)
				else:
					select = {'unit': 'T3StockSelector', 'config': {}}
					directive['select'] = select

				if 'config' not in select:
					select['config'] = {}

				if self.verbose:
					action = 'Modifying' if 'channel' in select['config'] else 'Adding'
					self.logger.log(VERBOSE,
						f'{action} channel selection criteria ({chan_name}) '
						f'for process {self.process["name"]}'
					)

				# processes embedded in channel must feature a transient selection.
				# An exception will be raised if someone embeds an admin t3 proc (without selection)
				# within a channel (unsupported feature)
				select['config']['channel'] = chan_name

		return self


	def apply_template(self) -> 'ProcessMorpher':
		""" Applies template possibly associated with process """

		# The process embedded in channel def requires templating itself
		if 'template' in self.process:

			if self.verbose:
				self.logger.log(VERBOSE,
					f'Applying template {self.process["template"]} '
					f'to process {self.process["name"]} '
					f'from distribution {self.process["distrib"]}'
				)

			self.process = self.templates[
				self.process['template']
			](**self.process).get_process(self.logger)

		return self


	def scope_aliases(self, first_pass_config: Dict) -> 'ProcessMorpher':
		""" Note: should be called before hash_t2_config """

		if self.process.get('distrib'):

			if self.verbose:
				self.logger.debug("Scoping aliases")

			recurse = False

			while walk_and_process_dict(
				arg = self.process,
				callback = self._scope_alias_callback,
				match = ['config'],
				first_pass_config = first_pass_config
			):
				if self.verbose:
					self.logger.debug("Alias(es) scoped %s" % ("again" if recurse else ""))
				recurse = True
		else:

			self.logger.error(
				f'Cannot scope aliases for process {self.process["name"]}'
				f' as it is missing a distribution name'
			)

		return self


	def _scope_alias_callback(self, path, k, d, **kwargs) -> bool:
		"""
		:returns: true if modification was done
		"""

		if 'first_pass_config' not in kwargs:
			raise ValueError('Parameter "first_pass_config" missing in kwargs')

		v = d[k]

		if v and isinstance(v, str):

			# Global alias
			if v[0] == '%':
				return False

			# Alias already scoped
			if "/" in v:
				return False

			scoped_alias = f'{self.process["distrib"]}/{v}'

			if not any([
				scoped_alias in kwargs['first_pass_config']['alias'][f't{tier}']
				for tier in (0, 1, 2, 3)
			]):
				raise ValueError(f'Alias "{scoped_alias}" not found')

			# Overwrite
			d[k] = scoped_alias

			if self.verbose:
				self.logger.log(VERBOSE, f'Config alias "{v}" renamed into "{scoped_alias}"')

			return True

		return False


	def resolve_aliases(self, t2d: Dict[str, Any], aliases: Dict[str, Any], root_path: str):
		"""
		Resolves aliases recursively (necessary before hashing t2 config)
		"""

		if self.verbose:
			self.logger.debug("Resolving aliases (required for hash)")

		recurse = False

		while walk_and_process_dict(
			arg = t2d,
			callback = self._resolve_alias_callback,
			match = ['config'],
			aliases = aliases,
			root_path = root_path
		):
			if self.verbose:
				self.logger.debug("Alias(es) resolved %s" % ("again" if recurse else ""))
			recurse = True


	def _resolve_alias_callback(self, path, k, d, **kwargs) -> None:
		"""
		Used by walk_and_process_dict(...) from resolve_aliases(...)
		"""

		if not isinstance(d[k], str):
			return

		aliases = kwargs.get('aliases')

		if not aliases:
			raise ValueError('Parameter "aliases" missing in kwargs')

		if d[k] not in aliases:
			raise ValueError(
				f'Unknown T2 config alias ({d[k]}) defined in process {self.process["name"]}'
			)

		if self.verbose:
			self.logger.debug(f"Resolving alias '{d[k]}' in {kwargs.get('root_path')}")

		d[k] = aliases[d[k]]


	def hash_t2_config(self, out_config: Dict) -> 'ProcessMorpher':
		"""
		This method modifies the underlying self._process dict structure.
		The 'config' (dict) value of UnitModel instances is replaced by a hash value (int).
		Notes:
		- Applies only to t0 and t1 processes
		- The config path filter "t2_compute.units" is used
		- Aliases are resolved (works recursively so that t2_dependency of tied units can use aliases as well)
		
		:param out_config: used to store new map entries in the ampel config: {<confid>: {<hash>: <conf dict>}}.
		"""

		if self.process['tier'] not in (0, 1):
			return self

		if self.verbose:
			self.logger.debug("Looking for t2 config to hash")

		# Example of conf_dicts keys
		# processor.config.directives.0.t0_add.t1_combine.0.t2_compute.units.0
 		# processor.config.directives.0.t0_add.t1_combine.0.t2_compute.units.0.config.t2_dependency.0
 		# processor.config.directives.0.t0_add.t2_compute.units.0
		conf_dicts: Dict[str, Dict[str, Any]] = {}

		walk_and_process_dict(
			arg = self.process,
			callback = self._gather_t2_config_callback,
			match = ['config'],
			conf_dicts = conf_dicts
		)

		# This does the trick of processing nested config first
		sorted_conf_dicts = {}
		for k in sorted(conf_dicts, key=len, reverse=True):
			sorted_conf_dicts[k] = conf_dicts[k]

		for k, d in sorted_conf_dicts.items():

			t2_unit = d["unit"]
			conf = d["config"]

			if self.verbose:
				extra = {'process': self.process['name'], 'conf': k + ".config"}
				self.logger.debug("Hashing T2 config", extra=extra)

			# Replace alias with content
			if isinstance(conf, str):
			
				if conf not in out_config['alias']['t2']:
					raise ValueError(
						f'Unknown T2 config alias ({conf}) defined in process {self.process["name"]}'
					)

				if self.verbose:
					self.logger.debug(f"Resolving alias '{conf}'", extra=extra)

				conf = out_config['alias']['t2'][conf]

			if isinstance(conf, dict):

				if override := d.get("override"):
					conf = {**conf, **override}

				if fqn := out_config['unit']['base'][t2_unit].get('fqn'):

					T2Unit = getattr(import_module(fqn), fqn.split('.')[-1])
					excl: Any = DataUnit._annots.keys()

					if issubclass(T2Unit, AbsPointT2Unit):
						excl = list(excl) + ["ingest"]

					if self.verbose:
						self.logger.debug("Creating model", extra=extra)

					# Model creation is required to take default config field values into account
					model = self._create_model(
						t2_unit,
						T2Unit._annots,
						T2Unit._defaults,
						excl,
					)

					conf = model(**conf).dict()

				else:

					self.logger.warn(
						f"T2 unit {t2_unit} not installed locally. "
						f"Building *unsafe* conf dict hash: "
						f"changes in unit defaults between releases will go undetected",
						extra=extra
					)

				if self.verbose:
					self.logger.debug("Computing hash", extra=extra)

				if conf is None:
					d["config"] = None
				else:
					d["config"] = out_config['confid'].add(conf)

			# For internal use only
			elif isinstance(conf, int):
				if conf not in out_config['confid']:
					raise ValueError(
						f'Unknown T2 config (int) alias defined in channel {k}:\n {t2_unit}'
					)
			else:
				raise ValueError(
					f'Unknown T2 config defined in channel {k}:\n {t2_unit}'
				)

		return self


	def _gather_t2_config_callback(self, path, k, d, **kwargs) -> None:
		"""
		Used by walk_and_process_dict(...) from hash_t2_config(...)
		"""

		if "t2_compute.units" in path:
			if self.verbose:
				self.logger.info("# path: %s.config" % path)
			kwargs['conf_dicts'][path] = d


	@classmethod
	def _create_model(cls, name, annotations, defaults, exclude):
		"""
		Build a pydantic model from annotations and defaults, replacing Secret
		fields with their dict representations
		"""
		fields = {
			k: (v, defaults[k] if k in defaults else ...)
			for k, v in annotations.items() if k not in exclude
		} # type: ignore
		# special case for Secret fields
		for k in list(fields.keys()):
			field_type = fields[k][0]
			if get_subtype(Secret, field_type):
				field_type = Dict[Literal["key"], str]
				if get_subtype(type(None), field_type):
					field_type = Optional[field_type]
				fields[k] = (field_type,) + fields[k][1:]
			elif type(field_type) is ModelMetaclass:
				fields[k] = (
					cls._create_model(
						field_type.__name__,
						field_type.__annotations__,
						field_type.__field_defaults__,
						set()
					),
					field_type.__field_defaults__
				)
		return create_model(
			name, __config__ = StrictModel.__config__,
			**fields
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsPointT2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 30.04.2020
# Last Modified Date: 30.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Union, Set, Tuple, List, Sequence, Optional
from ampel.base import abstractmethod
from ampel.type import DataPointId, ChannelId
from ampel.ingest.compile.CompilerBase import CompilerBase
from ampel.content.DataPoint import DataPoint

class AbsPointT2Compiler(CompilerBase, abstract=True):

	@abstractmethod
	def compile(self,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
		datapoints: Sequence[DataPoint]
	) -> Dict[Tuple[str, Optional[int], DataPointId], Set[ChannelId]]:
		pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t2/T2Processor.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 24.05.2019
# Last Modified Date: 02.04.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import gc
import signal
from time import time
from bson import ObjectId
from typing import Optional, List, Union, Type, Dict, Any, Sequence, Tuple, cast, overload

import backoff

from ampel.type import DataPointId, StockId, T, T2UnitResult, Tag
from ampel.db.query.utils import match_array
from ampel.util.collections import try_reduce
from ampel.log.utils import convert_dollars
from ampel.util.t1 import get_datapoint_ids
from ampel.enum.T2RunState import T2RunState
from ampel.enum.T2SysRunState import T2SysRunState
from ampel.struct.T2BroadUnitResult import T2BroadUnitResult
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.Compound import Compound
from ampel.content.T2Document import T2Document
from ampel.content.T2Record import T2Record
from ampel.content.JournalRecord import JournalRecord
from ampel.view.T2DocView import T2DocView, TYPE_POINT_T2, TYPE_STOCK_T2, TYPE_STATE_T2
from ampel.log import AmpelLogger, DBEventDoc, LogFlag, VERBOSE
from ampel.base.BadConfig import BadConfig
from ampel.log.utils import report_exception, report_error
from ampel.log.handlers.DefaultRecordBufferingHandler import DefaultRecordBufferingHandler
from ampel.util.mappings import build_unsafe_short_dict_id
from ampel.abstract.AbsProcessorUnit import AbsProcessorUnit
from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit
from ampel.abstract.AbsCustomStateT2Unit import AbsCustomStateT2Unit
from ampel.abstract.AbsTiedT2Unit import AbsTiedT2Unit
from ampel.abstract.AbsTiedPointT2Unit import AbsTiedPointT2Unit
from ampel.abstract.AbsTiedStateT2Unit import AbsTiedStateT2Unit
from ampel.abstract.AbsTiedStockT2Unit import AbsTiedStockT2Unit
from ampel.abstract.AbsTiedCustomStateT2Unit import AbsTiedCustomStateT2Unit
from ampel.model.UnitModel import UnitModel
from ampel.model.StateT2Dependency import StateT2Dependency
from ampel.core.JournalUpdater import JournalUpdater
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry

AbsT2 = Union[
	AbsStockT2Unit, AbsPointT2Unit, AbsStateT2Unit, AbsTiedPointT2Unit,
	AbsTiedStateT2Unit, AbsCustomStateT2Unit[T], AbsTiedCustomStateT2Unit[T],
]

stat_latency = AmpelMetricsRegistry.histogram(
	"latency",
	"Delay between T2 doc creation and processing",
	subsystem="t2",
	unit="seconds",
	labelnames=("unit", ),
)

stat_count = AmpelMetricsRegistry.counter(
	"docs_processed",
	"Number of T2 documents processed",
	subsystem="t2",
	labelnames=("unit", )
)

class T2Processor(AbsProcessorUnit):
	"""
	Fill results into pending :class:`T2 documents <ampel.content.T2Document.T2Document>`.
	"""

	#: ids of the t2 units to run. If not specified, any t2 unit will be run.
	t2_units: Optional[List[str]]
	#: process only those :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: with the given :attr:`~ampel.content.T2Document.T2Document.status`
	run_state: Union[T2SysRunState, Sequence[T2SysRunState]] = [
		T2SysRunState.NEW,
		T2SysRunState.NEW_PRIO,
		T2SysRunState.PENDING_DEPENDENCY,
		T2SysRunState.RERUN_REQUESTED
	]
	#: max number of t2 docs to process in :func:`run`
	doc_limit: Optional[int]
	#: tag(s) to add to the stock :class:`~ampel.content.JournalRecord.JournalRecord`
	#: every time a :class:`~ampel.content.T2Document.T2Document` is processed
	stock_jtag: Optional[Union[Tag, Sequence[Tag]]]
	#: create a "beacon" document indicating the last time T2Processor was run
	#: in the current configuration
	send_beacon: bool = True
	#: explicitly call :func:`gc.collect` after every document
	garbage_collect: bool = True
	#: process dependencies of :class:`tied T2 units <ampel.abtract.AbsTiedT2Unit.AbsTiedT2Unit>`
	run_dependent_t2s: bool = False
	#: maximum number of processing attempts per document
	max_try: int = 5


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self._ampel_db = self.context.db
		self._loader = self.context.loader

		# Ampel hashes unit class names in prod environment
		self.hashes: Dict[int, str] = {}
		self.optimize = self.context.config.get('general.optimize', int)

		if self.optimize and self.optimize > 1:
			for k, d in self.context.config.get('t2.unit.base', list): # type: ignore[union-attr]
				self.hashes[d['hash']] = k

		if isinstance(self.run_state, int):
			self.status_match: Union[T2SysRunState, Dict[str, List[T2SysRunState]]] = self.run_state
		elif isinstance(self.run_state, list):
			self.status_match = {"$in": self.run_state}
		else: # Could be a tuple (pymongo requires list)
			self.status_match = {"$in": list(self.run_state)}

		# Prepare DB query dict
		self.query: Dict[str, Any] = {'status': self.status_match}

		# Possibly restrict query to specified t2 units
		if self.t2_units:
			if len(self.t2_units) == 1:
				self.query['unit'] = self.t2_units[0]
			else:
				self.query['unit'] = {"$in": self.t2_units}

		# Shortcut
		self.col_stock = self._ampel_db.get_collection('stock')
		self.col_t0 = self._ampel_db.get_collection('t0')
		self.col_t1 = self._ampel_db.get_collection('t1')
		self.col_t2 = self._ampel_db.get_collection('t2')

		if self.send_beacon:
			self.create_beacon()

		# t2_instances stores unit instances so that they can be re-used in run()
		# Key: run config id(unit name + '_' + run_config name), Value: unit instance
		self.t2_instances: Dict[str, AbsT2] = {}

		self._run = True
		self._doc_counter = 0
		signal.signal(signal.SIGTERM, self.sig_exit)
		signal.signal(signal.SIGINT, self.sig_exit)


	def sig_exit(self, signum: int, frame) -> None:
		""" Executed when SIGTERM/SIGINT is caught. Stops doc processing in run() """
		self._run = False


	def create_beacon(self) -> None:
		""" Creates a beacon document if it does not exist yet """

		args = {
			'class': self.__class__.__name__,
			't2_units': self.t2_units,
			'run_state': self.status_match,
			'base_log_flag': self.base_log_flag.__int__(),
			'doc_limit': self.doc_limit
		}

		self.beacon_id = build_unsafe_short_dict_id(args)

		# Create beacon doc if it does not exist yet
		self.col_beacon = self._ampel_db.get_collection('beacon')

		# Create specific beacon doc if it does not exist yet
		self.col_beacon.update_one(
			{'_id': self.beacon_id},
			{
				'$setOnInsert': {
					'class': args.pop('class'),
					'config': convert_dollars(args)
				},
				'$set': {'timestamp': int(time())}
			},
			upsert=True
		)


	def run(self, pre_check: bool = True) -> int:
		""" :returns: number of t2 docs processed """

		# Add new doc in the 'events' collection
		event_doc = DBEventDoc(
			self._ampel_db, process_name=self.process_name, tier=2
		)

		# Avoid 'burning' a run_id for nothing (at the cost of a request)
		if pre_check and self.col_t2.find(self.query).count() == 0:
			return 0

		run_id = self.new_run_id()

		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, run_id,
			base_flag = LogFlag.T2 | LogFlag.CORE | self.base_log_flag
		)

		if self.send_beacon:
			self.col_beacon.update_one(
				{'_id': self.beacon_id},
				{'$set': {'timestamp': int(time())}}
			)

		jupdater = JournalUpdater(
			ampel_db = self.context.db, tier = 2, run_id = run_id,
			process_name = self.process_name, logger = logger,
			raise_exc = self.raise_exc, extra_tag = self.stock_jtag
		)

		# Loop variables
		self._doc_counter = 0
		update = {'$set': {'status': T2SysRunState.RUNNING}}
		doc_limit = self.doc_limit
		garbage_collect = self.garbage_collect

		# Process t2 docs until next() returns None (breaks condition below)
		self._run = True
		self._exception = False
		while self._run:

			# get t2 document (status is usually NEW or NEW_PRIO)
			t2_doc = self.col_t2.find_one_and_update(self.query, update)

			# Cursor exhausted
			if t2_doc is None:
				break
			elif logger.verbose > 1:
				logger.debug(f"T2 doc: {t2_doc}")

			self.process_t2_doc(t2_doc, logger, jupdater)

			# Check possibly defined doc_limit
			if doc_limit and self._doc_counter >= doc_limit:
				break

			if garbage_collect:
				gc.collect()

		event_doc.update(logger, docs=self._doc_counter, run=run_id, success=not self._exception)

		logger.flush()
		self.t2_instances.clear()
		return self._doc_counter


	def process_t2_doc(self,
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> T2Record: 

		before_run = time()

		t2_unit = self.get_unit_instance(t2_doc, logger)

		try:

			t2_rec: T2Record = {
				'run': jupdater.run_id,
			}
			j_rec: JournalRecord = {}

			if len(t2_doc.get("body", [])) > self.max_try:
				ret: T2UnitResult = T2SysRunState.TOO_MANY_TRIALS
			else:
				ret = self.run_t2_unit(t2_unit, t2_doc, logger, jupdater)

			# Used as timestamp and to compute duration below (using before_run)
			now = time()

			# _id is an ObjectId, but declared as bytes in ampel-interface to avoid
			# an explicit dependency on pymongo
			stat_latency.labels(t2_doc["unit"]).observe(
				now - cast("ObjectId", t2_doc["_id"]).generation_time.timestamp()
			)
			stat_count.labels(t2_doc["unit"]).inc()
			self._doc_counter += 1

			# New journal entry for the associated stock document
			j_rec = jupdater.add_record(
				stock = t2_doc['stock'],
				doc_id = t2_doc['_id'],
				unit = t2_doc['unit'],
				channel = try_reduce(t2_doc['channel'])
			)

			# New t2 sub-result entry (later appended with additional values)
			t2_rec = {
				'ts': j_rec['ts'],
				'run': jupdater.run_id,
				'duration': round(now - before_run, 3)
			}

			tag = None
			if v := t2_unit.get_version():
				t2_rec['version'] = v

			# T2 units can return a T2RunState integer rather than a dict instance / tuple
			# for example: T2RunState.ERROR, T2RunState.PENDING_DEPENDENCY, ...
			if isinstance(ret, int):
				if ret in T2SysRunState.__members__.values():
					logger.info(f'T2Processor status: {ret} ({T2SysRunState(ret).name})')
				else:
					logger.info(f'T2 unit returned int {ret}')
				doc_status = ret
				t2_rec['status'] = ret
				j_rec['status'] = ret

			# Unit returned just a dict result
			elif isinstance(ret, (dict, list)):
				doc_status = T2RunState.COMPLETED
				t2_rec['result'] = ret
				t2_rec['status'] = T2RunState.COMPLETED
				j_rec['status'] = T2RunState.COMPLETED

			# Customizations request by T2 unit
			elif isinstance(ret, T2BroadUnitResult):
				t2_rec['result'] = ret.rec_payload
				t2_rec['status'] = ret.rec_status
				doc_status = ret.doc_status
				tag = ret.doc_tag

				# This just sets a default as journal status can be customized via j_tweak
				j_rec['status'] = T2RunState.COMPLETED

				# Custom JournalTweak returned by unit
				if (j_tweak := ret.stock_journal_tweak) is not None:

					if j_tweak.tag:
						t2_rec['jup'] = True
						JournalUpdater.include_tags(j_rec, j_tweak.tag)

					if j_tweak.status:
						t2_rec['jup'] = True
						j_rec['status'] = j_tweak.status

					if j_tweak.extra:
						t2_rec['jup'] = True
						j_rec['extra'] = j_tweak.extra

			# Unsupported object returned by unit
			else:
				doc_status = t2_rec['status'] = j_rec['status'] = T2RunState.ERROR
				self._processing_error(
					logger, t2_doc=t2_doc, t2_rec=t2_rec, j_rec=j_rec,
					rec_msg='Unit returned invalid content',
					report_msg='Invalid content returned by T2 unit',
					extra={'ret': ret}
				)

			# TODO: check that unit did not use system reserved status

			# Empty payload returned, should not happen
			if 'result' in t2_rec and not t2_rec['result']:
				del t2_rec['result']
				logger.warn(
					"T2 unit return empty content",
					extra={'t2_doc': t2_doc}
				)

			self.push_t2_update(
				t2_doc['_id'], t2_rec, logger,
				tag=tag, status=doc_status
			)

			# Update stock document
			jupdater.flush()

		except Exception as e:
			if self.raise_exc:
				raise e
			self._processing_error(
				logger, t2_doc=t2_doc, t2_rec=t2_rec, j_rec=j_rec,
				exception=e, rec_msg='An exception occured'
			)

		return t2_rec


	@overload
	def load_input_docs(self,
		t2_unit: AbsPointT2Unit,
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> Union[None, T2SysRunState, DataPoint]:
		""
		...

	@overload
	def load_input_docs(self,
		t2_unit: AbsTiedPointT2Unit,
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> Union[None, T2SysRunState, Tuple[DataPoint, T2DocView]]:
		""
		...

	@overload
	def load_input_docs(self,
		t2_unit: AbsStateT2Unit,
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater,
	) -> Optional[Tuple[Compound, Sequence[DataPoint]]]:
		""
		...


	@overload
	def load_input_docs(self,
		t2_unit: AbsTiedStateT2Unit,
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> Union[None, T2SysRunState, Tuple[Compound, Sequence[DataPoint], List[T2DocView]]]:
		""
		...


	@overload
	def load_input_docs(self,
		t2_unit: AbsCustomStateT2Unit[T],
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> Optional[Tuple[T]]:
		""
		...


	@overload
	def load_input_docs(self,
		t2_unit: AbsTiedCustomStateT2Unit[T],
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> Union[None, T2SysRunState, Tuple[T, List[T2DocView]]]:
		""
		...

	@overload
	def load_input_docs(self,
		t2_unit: AbsStockT2Unit,
		t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater
	) -> Optional[Tuple["StockDocument"]]:
		...


	# NB: spell out union arg to ensure a common context for the TypeVar T
	def load_input_docs(self,
		t2_unit: AbsT2, t2_doc: T2Document,
		logger: AmpelLogger, jupdater: JournalUpdater
	) -> Union[
		None,
		T2SysRunState,                                         # Error / missing dependency
		DataPoint,                                             # point t2
		Tuple[DataPoint, T2DocView],                           # tied point t2
		Tuple["StockDocument"],                                # stock t2
		Tuple[Compound, Sequence[DataPoint]],                  # state t2
		Tuple[T],                                              # custom state t2 (T could be LightCurve)
		Tuple[Compound, Sequence[DataPoint], List[T2DocView]], # tied state t2
		Tuple[T, List[T2DocView]],                             # tied custom state t2
	]:
		"""
		Fetches documents required by `t2_unit`.

		:param t2_unit: instance to fetch input docs for
		:param t2_doc: document for which inputs are needed
		:param logger: logger
		:param jupdater: journal update service

		Regarding tied T2s (unit depends on other T2s):
		
		- Dependent T2s will be executed if run_dependent_t2s is True.
		- Note that the ingester ingests point and state t2s before state t2s so that the natural ordering
		  of T2 docs to be processed fits rather well a setup with a single T2processor instance processing all T2s.
		- For state t2s tied with other state t2s, putting the dependent units first in the AlertProcessor
		  directives will make sure these are ingested first and thus processed first by T2Processor.
		"""

		# State bound T2 units require loading of compound doc and datapoints
		if isinstance(
			t2_unit, (
				AbsStateT2Unit,
				AbsCustomStateT2Unit,
				AbsTiedStateT2Unit,
				AbsTiedCustomStateT2Unit
			)
		):
			datapoints: List[DataPoint] = []
			# If multiple links are avail, they should be all equivalent
			link = t2_doc['link'][0] if isinstance(t2_doc['link'], list) else t2_doc['link']
			compound: Optional[Compound] = self.load_compound(link)

			# compound doc must exist (None could mean an ingester bug)
			if compound is None:
				report_error(
					self._ampel_db, msg='Compound not found', logger=logger,
					info={'id': t2_doc['link'], 'doc': t2_doc}
				)
				return None

			# Datarights: suppress channel info (T3 uses instead a
			# 'projection' technic that should not be necessary here)
			compound.pop('channel')

			# Datapoints marked as excluded in the compound won't be included
			dps_ids = get_datapoint_ids(compound, logger)
			datapoints = self.load_datapoints(dps_ids)

			if not datapoints:
				report_error(
					self._ampel_db, msg='Datapoints not found', logger=logger,
					info={'id': compound, 'doc': t2_doc}
				)
				return None

			elif len(datapoints) != len(dps_ids):
				for el in set(dps_ids) - {el['_id'] for el in datapoints}:
					logger.error(f"Datapoint {el} referenced in compound not found")
				return None

			for dp in datapoints:
				dp.pop('excl', None) # Channel based exclusion
				dp.pop('extra', None)
				dp.pop('policy', None)

			if isinstance(t2_unit, AbsTiedT2Unit):

				queries: List[Dict[str, Any]] = []

				if isinstance(t2_unit, AbsTiedCustomStateT2Unit):
					# A LightCurve instance for example
					custom_state = t2_unit.build(compound, datapoints)

				for tied_model in t2_unit.get_t2_dependencies():

					d = self.build_tied_t2_query(t2_unit, tied_model, t2_doc)
					if d['link'] is None:

						if isinstance(tied_model, StateT2Dependency) and tied_model.link_override:
							if isinstance(t2_unit, AbsTiedCustomStateT2Unit):
								d['link'] = t2_unit.get_link(tied_model.link_override, custom_state)
							else:
								d['link'] = t2_unit.get_link( # type: ignore[union-attr] # mypy forgot the first "if" of this method
									tied_model.link_override, compound, datapoints
								)
						else:
							d['link'] = match_array(dps_ids)

					queries.append(d)

				tied_views = self.run_tied_queries(queries, t2_doc, jupdater, logger)

				# Dependency missing
				if isinstance(tied_views, T2SysRunState):
					return tied_views

				if isinstance(t2_unit, AbsTiedStateT2Unit):
					return (compound, datapoints, tied_views)

				# instance of AbsTiedCustomStateT2Unit
				return (custom_state, tied_views)

			else:

				if isinstance(t2_unit, AbsStateT2Unit):
					return (compound, datapoints)

				else: # instance of AbsCustomStateT2Unit
					return (t2_unit.build(compound, datapoints), )

		elif isinstance(t2_unit, (AbsStockT2Unit, AbsPointT2Unit, AbsTiedPointT2Unit)):

			if doc := (
				self.load_datapoint(t2_doc['link'])
				if isinstance(t2_unit, AbsPointT2Unit)
				else self.load_stock(t2_doc['link'])
			):
				if isinstance(t2_unit, AbsTiedPointT2Unit):

					tied_views = self.run_tied_queries(
						[
							self.build_tied_t2_query(t2_unit, tied_model, t2_doc)
							for tied_model in t2_unit.get_t2_dependencies()
						],
						t2_doc, jupdater, logger
					)

					if isinstance(tied_views, T2SysRunState):
						return tied_views

					return (doc, tied_views)

				return (doc, )

			report_error(
				self._ampel_db, msg='Datapoint not found' if isinstance(t2_unit, (AbsTiedPointT2Unit, AbsPointT2Unit))
				else 'Stock doc not found', logger=logger, info={'doc': t2_doc}
			)

			return None

		else:

			report_error(
				self._ampel_db, msg='Unknown T2 unit type',
				logger=logger, info={'doc': t2_doc}
			)

			return None


	# allow 10 seconds for input documents to become available
	@backoff.on_predicate(backoff.expo, max_time=10)
	def load_compound(self, link: bytes) -> Optional[Compound]:
		return next(self.col_t1.find({'_id': link}), None)


	@backoff.on_predicate(backoff.expo, max_time=10)
	def load_datapoints(self, dps_ids: List[DataPointId]) -> List[DataPoint]:
		return dps if len(dps := list(self.col_t0.find({'_id': {"$in": dps_ids}}))) == len(dps_ids) else list()


	@backoff.on_predicate(backoff.expo, max_time=10)
	def load_datapoint(self, link: DataPointId) -> Optional[DataPoint]:
		return next(self.col_t0.find({'_id': link}), None)


	@backoff.on_predicate(backoff.expo, max_time=10)
	def load_stock(self, link: StockId) -> Optional[StockDocument]:
		return next(self.col_stock.find({'_id': link}), None)


	def run_tied_queries(self,
		queries: List[Dict[str, Any]],
		t2_doc: T2Document,
		jupdater: JournalUpdater,
		logger: AmpelLogger
	) -> Union[T2SysRunState, List[T2DocView]]:

		t2_views: List[T2DocView] = []

		for query in queries:

			if self.run_dependent_t2s:

				processed_ids: List[ObjectId] = []

				# run pending dependencies
				while (
					dep_t2_doc := self.col_t2.find_one_and_update(
						{'status': self.status_match, **query},
						{'$set': {'status': T2SysRunState.RUNNING}}
					)
				) is not None:

					if not dep_t2_doc.get('body'):
						dep_t2_doc['body'] = []

					dep_t2_doc['body'].append(
						self.process_t2_doc(dep_t2_doc, logger, jupdater)
					)

					# suppress channel info
					dep_t2_doc.pop('channel')
					t2_views.append(self.view_from_record(dep_t2_doc))
					processed_ids.append(dep_t2_doc['_id'])

				if len(processed_ids) > 0:
					query['_id'] = {'$nin': processed_ids}

			# collect dependencies
			for dep_t2_doc in self.col_t2.find(query):
				# suppress channel info
				dep_t2_doc.pop('channel')
				t2_views.append(self.view_from_record(dep_t2_doc))

		if not self.run_dependent_t2s:
			for view in t2_views:
				if view.get_payload() is None:
					if logger.verbose > 1:
						logger.debug("Dependent T2 unit not run yet", extra={'t2_oid': t2_doc['_id']})
					return T2SysRunState.PENDING_DEPENDENCY

		return t2_views


	def view_from_record(self, doc: T2Document) -> T2DocView:
		"""
		We might want to move this method elsewhere in the future
		"""

		t2_unit_info = self.context.config.get(f'unit.base.{doc["unit"]}', dict)
		if not t2_unit_info:
			raise ValueError(f'Unknown T2 unit {doc["unit"]}')

		if 'AbsStockT2Unit' in t2_unit_info['base']:
			t2_type: int = TYPE_STOCK_T2
		elif 'AbsPointT2Unit' in t2_unit_info['base']:
			t2_type = TYPE_POINT_T2
		else: # quick n dirty
			t2_type = TYPE_STATE_T2

		return T2DocView(
			stock = doc['stock'],
			unit = doc['unit'],
			link = doc['link'],
			status = doc['status'],
			t2_type = t2_type,
			created = doc['_id'].generation_time.timestamp(),
			body = doc.get('body'),
			config = self.context.config.get(
				f'confid.{doc["config"]}', dict
			) if doc['config'] else None
		)


	def build_tied_t2_query(self,
		t2_unit: AbsT2, tied_model: UnitModel, t2_doc: T2Document
	) -> Dict[str, Any]:
		"""
		This method handles "default" situations.
		Callers must check returned 'link'.
		If None (state t2 linked with point t2), further handling is required

		:raises:
			- ValueError if functionality is not implemented yet
			- BadConfig if what's requested is not possible (a point T2 cannot be linked with a state t2)
	
		:returns: link value to match or None if further handling is required (state t2 linked with point t2)
		"""

		t2_unit_info = self.context.config.get(
			f'unit.base.{tied_model.unit}', dict
		)

		if not t2_unit_info:
			raise ValueError(f'Unknown T2 unit {tied_model.unit}')

		d: Dict[str, Any] = {
			'unit': tied_model.unit,
			'config': tied_model.config,
			'channel': {"$in": t2_doc['channel']},
			'stock': t2_doc['stock'],
		}

		if isinstance(t2_unit, AbsTiedPointT2Unit):

			if 'AbsPointT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['link']

			elif 'AbsStockT2Unit' in t2_unit_info['base']:
				raise ValueError('Not implemented yet')

			else: # State T2
				raise BadConfig('Tied point T2 cannot be linked with state T2s')

		elif isinstance(t2_unit, AbsTiedStockT2Unit):

			if 'AbsPointT2Unit' in t2_unit_info['base']:
				raise BadConfig('Tied stock T2 cannot be linked with point T2s')

			elif 'AbsStockT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['stock']

			else: # State T2
				raise BadConfig('Tied stock T2 cannot be linked with state T2s')

		if isinstance(t2_unit, (AbsTiedStateT2Unit, AbsTiedCustomStateT2Unit)):

			if 'AbsPointT2Unit' in t2_unit_info['base']:
				d['link'] = None # Further checks required (link_override check)

			elif 'AbsStockT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['stock']

			else: # State T2
				d['link'] = t2_doc['link']

		return d


	def run_t2_unit(self,
		t2_unit: AbsT2, t2_doc: T2Document, logger: AmpelLogger, jupdater: JournalUpdater,
	) -> T2UnitResult:
		"""
		Regarding the possible int return code:
		usually, if an int is returned, it should be a T2RunState member
		but let's not be too restrictive here
		"""

		args = self.load_input_docs(t2_unit, t2_doc, logger, jupdater)

		if args is None:
			return T2RunState.ERROR

		if isinstance(args, int):
			return args

		ret = t2_unit.run(*args)

		if t2_unit._buf_hdlr.buffer: # type: ignore[union-attr]
			t2_unit._buf_hdlr.forward( # type: ignore[union-attr]
				logger, stock=t2_doc['stock'], channel=t2_doc['channel']
			)

		return ret


	def _processing_error(self,
		logger: AmpelLogger, t2_doc: T2Document, t2_rec: T2Record,
		j_rec: JournalRecord, rec_msg: str, report_msg: Optional[str] = None,
		extra: Dict[str, Any] = {}, exception: Optional[Exception] = None
	) -> None:
		"""
		- Updates the t2 document by appending a T2Record and
		updating 'status' with value T2RunState.EXCEPTION if an exception
		if provided through the parameter 'exception' or with value
		T2RunState.ERROR otherwise. The field  'msg' of the created
		T2Record entry will be set to the value of parameter 'rec_msg'.

		- Updates the stock document by appending a JournalRecord to it
		and by updating the 'modified' timestamp

		- Creates a 'trouble' document in the troubles collection
		to report the incident
		"""

		t2_rec = t2_rec.copy()
		t2_rec['msg'] = rec_msg

		self.push_t2_update(
			t2_doc['_id'], t2_rec, logger,
			status = T2SysRunState.EXCEPTION if exception else T2RunState.ERROR
		)

		info: Dict[str, Any] = {
			**extra,
			'process': self.process_name,
			'run': t2_rec['run'],
			'stock': t2_doc['stock'],
			'doc': t2_doc
		}

		if exception:
			self._exception = True
			report_exception(
				self._ampel_db, logger=logger,
				exc=exception, info=info
			)
		else:
			report_error(
				self._ampel_db, logger=logger,
				msg=report_msg if report_msg else '', info=info
			)


	def push_t2_update(self,
		t2_id: ObjectId, t2_rec: T2Record, logger: AmpelLogger, *,
		tag: Optional[Union[Tag, List[Tag]]] = None, status: int = T2RunState.COMPLETED
	) -> None:
		""" Performs DB updates of the T2 doc and stock journal """

		if logger.verbose:
			logger.log(VERBOSE, 'Saving T2 unit result')

		setd: Dict[str, Any] = {'status': status}
		if tag:
			setd['$addToSet'] = {'$each': tag} if isinstance(tag, list) else tag

		# Update T2 document
		self.col_t2.update_one(
			{'_id': t2_id},
			{
				'$push': {'body': t2_rec},
				'$set': setd
			}
		)


	def get_unit_instance(self, t2_doc: T2Document, logger: AmpelLogger) -> AbsT2:

		k = f'{t2_doc["unit"]}_{t2_doc["config"]}'

		# Check if T2 instance exists in this run
		if k not in self.t2_instances:

			if 'col' in t2_doc:

				if 't0' in t2_doc['col']:  # type: ignore[operator] # mypy inference issue
					sub_type: Type[AbsT2] = AbsPointT2Unit
				elif 'stock' in t2_doc['col']: # type: ignore[operator] # mypy inference issue
					sub_type = AbsStockT2Unit
				else:
					raise ValueError('Unsupported t2 unit type')

			else:

				bcs = self.context.config.get( # type: ignore[call-overload] # improve AmpelConfig overloads
					f"unit.base.{t2_doc['unit']}.base", (list, tuple)
				)

				if bcs is None:
					raise ValueError(f'Unknown t2 unit: {t2_doc["unit"]}')

				if "AbsStateT2Unit" in bcs:
					sub_type = AbsStateT2Unit
				elif "AbsCustomStateT2Unit" in bcs:
					sub_type = AbsCustomStateT2Unit
				elif "AbsTiedStateT2Unit" in bcs:
					sub_type = AbsTiedStateT2Unit
				elif "AbsTiedCustomStateT2Unit" in bcs:
					sub_type = AbsTiedCustomStateT2Unit
				else:
					raise ValueError(f'Unsupported t2 unit (base classes: {bcs})')

			# Create channel (buffering) logger
			buf_hdlr = DefaultRecordBufferingHandler(level=logger.level)
			buf_logger = AmpelLogger.get_logger(
				name = k,
				base_flag = (getattr(logger, 'base_flag', 0) & ~LogFlag.CORE) | LogFlag.UNIT,
				console = False,
				handlers = [buf_hdlr]
			)

			# Instantiate t2 unit
			unit_instance = self._loader.new_base_unit(
				unit_model = UnitModel(
					unit = t2_doc['unit'] if isinstance(t2_doc['unit'], str)
						else self.hashes[t2_doc['unit']],
					config = t2_doc['config']
				),
				logger = buf_logger,
				sub_type = sub_type
			)

			# Shortcut to avoid unit_instance.logger.handlers[?]
			setattr(unit_instance, '_buf_hdlr', buf_hdlr)
			self.t2_instances[k] = unit_instance # type: ignore[assignment] # it's fine

		return self.t2_instances[k]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/utils/crypto.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.04.2020
# Last Modified Date: 18.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import hashlib, sys, collections
from typing import Dict, Type, TypeVar, Literal, Union, Any, Optional, List, Iterable
from ampel.model.EncryptedDataModel import EncryptedDataModel

HT = TypeVar("HT", bytes, str, int)


def b2_short_hash(payload: Union[bytes, str]) -> int:
	if isinstance(payload, str):
		payload = bytes(payload, "utf8")
	return hash_payload(payload, int, 'blake2b', digest_size=7)


def hash_payload(
	payload: bytes,
	ret: Type[HT] = bytes, # type: ignore[assignment]
	alg: Literal['sha512', 'blake2b', 'sha1', 'md5'] = 'sha512',
	**kwargs
) -> HT:
	"""
	:param ret: return type, can be bytes, str (hex digest) or int
	:param alg: hash algorithm (default is sha512)
	:param kwargs: forwarded to hashlib hash function
	"""

	# Create hash object
	ho = getattr(hashlib, alg)(payload, **kwargs)

	# using issubclass and not == because of mypy
	if issubclass(ret, bytes):
		return ho.digest()

	if issubclass(ret, int):
		return int.from_bytes(
			ho.digest(),
			byteorder=sys.byteorder
		)

	return ho.hexdigest()


def aes_recursive_decrypt(d: Dict[str, Any], pwds: Iterable[str], debug: bool = False) -> Optional[dict]:
	"""
	Note: all dicts are deep copied. Lists are copied only if they contain encrypted dicts.
	:returns: the conf entry with decrypted entries when applicable

	example:

	In []: d = {
		"mongo": {
			"writer": "mongodb://localhost:27017",
			"logger": "mongodb://localhost:27017"
		},
		"ampel-contrib-hu/slack": {
			"token": {
				"iv": "KRbaS6FY3zV+OanX2Vszxg==",
				"v": 1,
				"iter": 1000,
				"ks": 128,
				"ts": 64,
				"mode": "ccm",
				"adata": "",
				"cipher": "aes",
				"salt": "Xzur+GnyJFs=",
				"ct": "Z0OmujCT0tnubZ3wY6bX"
			},
			"channel": "#too-metachannel"
		}
	}

	In []: aes_recursive_decrypt(d, ["abc", "def", "yoyo"])
	Out[3]: {
		"mongo": {
			"writer": "mongodb://localhost:27017",
			"logger": "mongodb://localhost:27017"
		},
		"ampel-contrib-hu/slack": {
 			'token': 'bla bla',
			'channel': '#too-metachannel'
		}
	}
	"""

	if not isinstance(d, dict):
		return None

	ret: Dict[str, Any] = {}

	for k, v in d.items():

		if isinstance(v, dict):

			if 'iv' in v and 'ks' in v:

				try:
					edm = EncryptedDataModel(**v)
					ret[k] = edm.decrypt(pwds)
				except Exception:
					if debug:
						import traceback
						print("#" * 50)
						traceback.print_exc()
						print("#" * 50)
					ret[k] = v
			else:
				ret[k] = aes_recursive_decrypt(v, pwds)

			continue

		if isinstance(v, collections.abc.Sequence):
			if any(isinstance(el, dict) and 'iv' in el for el in v):
				ret[k] = [
					aes_recursive_decrypt(el, pwds) if isinstance(el, dict) else el
					for el in v
				]
				continue

		ret[k] = v

	return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsApplicable.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 27.02.2020
# Last Modified Date: 09.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any
from ampel.base import abstractmethod, AmpelBaseModel
from ampel.base.AmpelABC import AmpelABC

class AbsApplicable(AmpelABC, AmpelBaseModel, abstract=True):

	@abstractmethod
	def apply(self, arg: Any) -> Any:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/compile/StateT2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.05.2020
# Last Modified Date: 01.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.abstract.ingest.AbsStateT2Compiler import AbsStateT2Compiler


class StateT2Compiler(AbsStateT2Compiler):
	"""
	Helper class capabable of generating a nested dict that is used as basis to create T2 documents.
	The generated structure is optimized: multiple T2 documents associated with the same stock record
	accross different channels are merged into one single T2 document
	that references all corrsponding channels.
	"""

	# Not implemented yet
	pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/PointT2Ingester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 23.03.2020
# Last Modified Date: 11.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from pymongo import UpdateOne
from typing import Sequence, Dict, Optional, Union, Literal, Tuple, List
from ampel.type import StockId, ChannelId
from ampel.enum.T2SysRunState import T2SysRunState
from ampel.content.DataPoint import DataPoint
from ampel.ingest.compile.PointT2Compiler import PointT2Compiler
from ampel.abstract.ingest.AbsT2Ingester import AbsT2Ingester
from ampel.abstract.ingest.AbsPointT2Ingester import AbsPointT2Ingester
from ampel.abstract.ingest.AbsPointT2Compiler import AbsPointT2Compiler


class PointT2Ingester(AbsPointT2Ingester):

	compiler: AbsPointT2Compiler = PointT2Compiler()
	default_ingest_config: Dict[
		Literal['eligible'],
		Optional[Union[Literal['first', 'last', 'all'], Tuple[int, int, int]]]
	] = {"eligible": None} # None means all eligible


	def ingest(self,
		stock_id: StockId,
		datapoints: Sequence[DataPoint],
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]]
	) -> None:

		optimized_t2s = self.compiler.compile(chan_selection, datapoints)
		now = int(time())

		# Loop over t2 units to be created
		for (t2_id, run_config, link_id), chans in optimized_t2s.items():

			jchan, chan_add_to_set = AbsT2Ingester.build_query_parts(chans)

			# Append update operation to bulk list
			self.updates_buffer.add_t2_update(
				UpdateOne(
					# Matching search criteria
					{
						'stock': stock_id,
						'unit': t2_id,
						'config': run_config,
						'link': link_id,
						'col': 't0'
					},
					{
						# Attributes set if no previous doc exists
						'$setOnInsert': {
							'stock': stock_id,
							'tag': self.tags,
							'unit': t2_id,
							'link': link_id,
							'config': run_config,
							'status': T2SysRunState.NEW.value,
							'col': 't0'
						},
						# Journal and channel update
						'$addToSet': {
							'channel': chan_add_to_set,
							'journal': {
								'tier': self.tier,
								'dt': now,
								'channel': jchan
							}
						}
					},
					upsert=True
				)
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/IngestionDirective.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 05.06.2020
# Last Modified Date: 10.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pydantic import validator
from typing import Union, Optional, List
from ampel.type import ChannelId
from ampel.model.UnitModel import UnitModel
from ampel.model.StrictModel import StrictModel
from ampel.model.ingest.T0AddModel import T0AddModel
from ampel.model.ingest.T1CombineModel import T1CombineModel
from ampel.model.ingest.T2ComputeModel import T2ComputeModel


class IngestionDirective(StrictModel):

	#: Channel for which to create documents
	channel: ChannelId
	#: Add new :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	t0_add: Optional[T0AddModel]
	#: Create :class:`compounds <ampel.content.Compound.Compound>` from
	#: ingested :class:`datapoints <ampel.content.DataPoint.DataPoint>` and the
	#: :class:`stock <ampel.content.StockDocument.StockDocument>`, e.g. drawing
	#: from other data streams
	t1_combine: Optional[List[T1CombineModel]]
	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>` bound to :class:`stocks <ampel.content.StockDocument.StockDocument>`
	t2_compute: Optional[T2ComputeModel]
	#: Update the :class:`stock <ampel.content.StockDocument.StockDocument>`
	stock_update: Union[UnitModel, str]

	@validator('stock_update', pre=True)
	def _allow_str(cls, v):
		if isinstance(v, str):
			return UnitModel(unit=v)
		return v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/template/PeriodicSummaryT3.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 10.08.2020
# Last Modified Date: 10.08.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from logging import Logger
from typing import Any, Dict, Literal, Optional, Sequence, Union, List

from ampel.abstract.AbsProcessTemplate import AbsProcessTemplate
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.StrictModel import StrictModel
from ampel.model.t3.LoaderDirective import LoaderDirective
from ampel.model.t3.T2FilterModel import T2FilterModel
from ampel.model.UnitModel import UnitModel
from ampel.type import ChannelId, Tag


class FilterModel(StrictModel):
    #: Filter based on T2 results
    t2: Union[T2FilterModel, AllOf[T2FilterModel], AnyOf[T2FilterModel]]

UnitModelOrString = Union[UnitModel, str]
UnitModelSequence = Union[Sequence[UnitModelOrString], UnitModelOrString]

class PeriodicSummaryT3(AbsProcessTemplate):
    """
    A T3 process that selects stocks modified since its last invocation, and
    supplies them, to a sequence of AbsT3Units.
    """

    #: Process name.
    name: str
    tier: Literal[3] = 3
    active: bool = True
    #: one or more `schedule <https://schedule.readthedocs.io/en/stable/>`_
    #: expressions, e.g: ``every().day.at("15:00")`` or ``every(42).minutes``
    #: 
    #: .. note:: all times are are expressed in UTC
    schedule: Union[str, Sequence[str]]
    #: Channel selection.
    channel: Union[
        None, ChannelId, AllOf[ChannelId], AnyOf[ChannelId], OneOf[ChannelId]
    ] = None
    distrib: Optional[str]
    source: Optional[str]
    #: Stock tag selection.
    tag: Optional[
        Dict[
            Literal["with", "without"],
            Union[Tag, Dict, AllOf[Tag], AnyOf[Tag], OneOf[Tag]],
        ]
    ] = None
    #: Documents to load. If a string, should refer to an entry in the
    #: ``alias.t3`` config section. See :ref:`t3-directive-load`.
    load: Optional[Sequence[Union[str,LoaderDirective]]] = None
    #: Additional stock filters.
    filter: Optional[FilterModel] = None
    #: Complement stages. See :ref:`t3-directive-complement`.
    complement: Optional[UnitModelSequence] = None
    #: Units to run. See :ref:`t3-directive-run-execute`.
    run: UnitModelSequence

    def get_process(self, logger: Logger) -> Dict[str, Any]:
        directive: Dict[str, Any] = {
            "context": [
                {"unit": "T3AddAlertsNumber"},
            ],
            "select": {
                "unit": "T3StockSelector",
                "config": {
                    "modified": {
                        "after": {
                            "match_type": "time_last_run",
                            "process_name": self.name,
                        },
                        "before": {"match_type": "time_delta"},
                    },
                    "channel": self.channel,
                    "tag": self.tag,
                },
            },
            "load": {"unit": "T3SimpleDataLoader", "config": {}},
            "run": {
                "unit": "T3UnitRunner",
                "config": {
                    "directives": [{"execute": self.get_units(self.run)}]
                },
            },
        }

        # Restrict stock selection according to T2 values
        if self.filter:
            directive["select"]["unit"] = "T3FilteringStockSelector"
            directive["select"]["config"]["t2_filter"] = self.filter.t2.dict()

        if self.channel is not None:
            # load only documents that pass channel selection
            directive["load"]["config"]["channel"] = self.channel
            # project subdocuments (e.g. stock journal) to channel selection
            directive["run"]["config"]["directives"][0]["project"] = {
                "unit": "T3ChannelProjector",
                "config": {"channel": self.channel},
            }
    
        # Restrict document types to load
        if self.load is not None:
            directive["load"]["config"]["directives"] = self.load
        else:
            del directive["load"]

        if self.complement is not None:
            directive["complement"] = self.get_units(self.complement)

        ret: Dict[str, Any] = {
            "tier": self.tier,
            "schedule": self.schedule,
            "active": self.active,
            "distrib": self.distrib,
            "source": self.source,
            "channel": self.get_channel_tag(),
            "name": self.name,
            "processor": {
                "unit": "T3Processor",
                "config": {"directives": [directive]},
            },
        }

        return self._to_dict(ret)

    @classmethod
    def _to_dict(cls, item):
        if isinstance(item, dict):
            return {k: cls._to_dict(v) for k, v in item.items()}
        elif isinstance(item, list):
            return [cls._to_dict(v) for v in item]
        elif hasattr(item, "dict"):
            return cls._to_dict(item.dict())
        else:
            return item

    def get_units(self, units: UnitModelSequence) -> List[Dict[str,Any]]:
        if isinstance(units, str):
            return [UnitModel(unit=units).dict()]
        elif isinstance(units, UnitModel):
            return [units.dict()]
        else:
            return [self.get_units(u)[0] for u in units]

    def get_schedule(self) -> Sequence[str]:
        if isinstance(self.schedule, str):
            return [self.schedule]
        else:
            return self.schedule

    def get_channel_tag(self) -> Union[None, str, int]:
        """
        Get channel if single channel, otherwise None
        """
        if isinstance(self.channel, str) or isinstance(self.channel, int):
            return self.channel
        else:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsT2Ingester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 23.03.2020
# Last Modified Date: 19.11.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

import importlib
from typing import Sequence, Dict, Any, Optional, Union, List, Literal, Set, Tuple
from ampel.type import ChannelId
from ampel.base.DataUnit import DataUnit
from ampel.abstract.ingest.AbsIngester import AbsIngester
from ampel.abstract.ingest.AbsStateT2Compiler import AbsStateT2Compiler
from ampel.abstract.ingest.AbsPointT2Compiler import AbsPointT2Compiler
from ampel.abstract.ingest.AbsStockT2Compiler import AbsStockT2Compiler
from ampel.model.ingest.T2IngestModel import T2IngestModel


class AbsT2Ingester(AbsIngester, abstract=True):

	compiler: Union[AbsStateT2Compiler, AbsPointT2Compiler, AbsStockT2Compiler]
	tags: Optional[List[Union[int, str]]]
	tier: Literal[0, 1, 3] = 0
	default_ingest_config: Dict = {}


	def add_ingest_models(self, channel: ChannelId, models: Sequence[T2IngestModel]):
		""" This method is typically called multiple times by the AP """

		for im in models:

			T2Class = None
			if isinstance(im.unit, str):
				unit_name = im.unit
				t2_info = self.context.config.get(f'unit.base.{unit_name}', dict)

				if not t2_info:
					raise ValueError(f'Unknown T2 unit {unit_name}')

				# The unit is installed locally
				if 'fqn' in t2_info:
					T2Class = self.context.loader.get_class_by_name(
						name = unit_name, unit_type = DataUnit
					)

				for el in t2_info['base']:
					# Pickup first Abstract class in mro()
					# Note: this strategy might cause pblms
					if el.startswith("Abs"):
						try:
							T2AbsClass = getattr(
								# import using fully qualified name
								importlib.import_module(f"ampel.abstract.{el}"), el
							)
						except Exception:
							raise ValueError(f"Unknown abstract class: {el}")

				if "T2AbsClass" not in locals(): # this should be unlikely to happen
					raise ValueError(f"No abstract class found for {unit_name}")
			else:
				T2Class = im.unit
				for base in T2Class.__mro__:
					if base.__name__.startswith("Abs"):
						T2AbsClass = base
						break
				else:
					raise ValueError(f"No abstract class found for {T2Class.__name__}")

			# Ingest options prevalence/priority from low to high:
			# 1) Default options from ingester (ex: T2PhotoIngester might default upper_limits to False)
			# 2) Default options from last T2 Abstract class (ex: AbsLightCurveT2Unit might default upper_limits to True)
			# 3) Default options from T2 class (ex: T2SNCosmo might default upper_limits to False)
			# 4) Specific options from model config (t2_compute)
			# (ex: custom T2SNCosmo ticket might request upper_limits nonetheless)
			ingest_options: Dict[str, Any] = {
				**self.default_ingest_config,
				**getattr(T2AbsClass, 'ingest', {}),
				**getattr(T2Class, 'ingest', {}),
				**(im.ingest if im.ingest else {})
			}

			self.compiler.add_ingest_model(channel, im)
			self.compiler.set_ingest_options(channel, im, ingest_options)


	@staticmethod
	def build_query_parts(chan_set: Set[ChannelId]) -> Tuple[
		Union[ChannelId, List[ChannelId]],
		Union[ChannelId, Dict[Literal['$each'], List[ChannelId]]]
	]:
		"""
		First tuple member is for the journal 'channel' field
		Second tuple member if for the operation $addToSet on root 'channel' field
		"""
		if len(chan_set) == 1:
			chan = next(iter(chan_set))
			return chan, chan
		else:
			chans = list(chan_set) # pymongo requires list
			return chans, {'$each': chans}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/ChannelConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 06.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional
from ampel.util.crypto import b2_short_hash
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class ChannelConfigCollector(AbsDictConfigCollector):

	def add(self,
		arg: Dict[str, Any],
		file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> None:

		if 'channel' not in arg:
			self.missing_key('channel', 'channel', file_name, dist_name)
			return

		try:

			chan_name = arg['channel']

			if self.verbose:
				self.logger.log(VERBOSE, f'Adding channel: {chan_name}')

			if 'distrib' in arg:
				dist_name = arg['distrib']
			else:
				if dist_name:
					arg['distrib'] = dist_name

			if 'source' in arg:
				file_name = arg['source']
			else:
				if file_name:
					arg['source'] = file_name

			# Check duplicated channel names
			if self.get(chan_name):
				self.duplicated_entry(
					conf_key = chan_name,
					new_file = file_name,
					new_dist = dist_name,
					prev_file = self.get(chan_name).get('conf', 'unknown'), # type: ignore
					prev_dist = self.get(chan_name).get('distrib', 'unknown') # type: ignore
				)
				return

			if not ('NO_HASH' in arg.get('policy', []) or isinstance(chan_name, int)):
				arg['hash'] = b2_short_hash(chan_name)
				for k, v in self.items():
					if arg['hash'] == v.get('hash'):
						raise ValueError(
							f'Channel name hash collision detected. '
							f'Channel name 1: {k}. Hash value: {v.get("hash")}'
							f'Channel name 2: {arg["channel"]}. Hash value: {arg.get("hash")}'
						)

			self.__setitem__(chan_name, arg)

		except Exception as e:
			self.error(
				f'Error occured while loading channel config {self.distrib_hint(file_name, dist_name)}. '
				f'Offending value: {arg}',
				exc_info=e
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsCompoundIngester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 08.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from logging import WARNING
from typing import Sequence, Generic, List, Tuple, Union, Optional
from ampel.base import abstractmethod
from ampel.content.DataPoint import DataPoint
from ampel.abstract.ingest.AbsIngester import AbsIngester
from ampel.type import StockId, TypeVar, ChannelId
from ampel.ingest.CompoundBluePrint import CompoundBluePrint
from ampel.log.handlers.RecordBufferingHandler import RecordBufferingHandler

T = TypeVar("T", bound=CompoundBluePrint)


class AbsCompoundIngester(Generic[T], AbsIngester, abstract=True):


	@abstractmethod
	def ingest(self,
		stock_id: StockId,
		datapoints: Sequence[DataPoint],
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]]
	) -> Optional[T]:
		...


	@abstractmethod
	def add_channel(self, channel: ChannelId) -> None:
		...


	def log_records_to_logd(self,
		rec_buffer: RecordBufferingHandler, clear: bool = True
	) -> None:
		"""
		Compound ingesters can make use of a T1 unit underneath,
		which requires a logger like every other base units.
		The logger used in usually associated with a buffering handler
		whose logs are later transfered to the LogsBufferDict instance
		used and shared among ingesters.
		"""

		logd = self.logd
		for rec in rec_buffer.buffer:
			if rec.msg:
				logd['logs'].append(rec.msg) # type: ignore
			if rec.levelno > WARNING:
				logd['err'] = True
			if hasattr(rec, 'extra'):
				for k, v in rec.extra: # type: ignore
					logd['extra'][k] = v

		if clear:
			rec_buffer.buffer = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsDataPointIngester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 18.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Sequence, Generic, Any
from ampel.base import abstractmethod
from ampel.content.DataPoint import DataPoint
from ampel.abstract.ingest.AbsIngester import AbsIngester
from ampel.type import StockId, TypeVar

T = TypeVar("T", bound=DataPoint)


class AbsDataPointIngester(Generic[T], AbsIngester, abstract=True):
	""" No known implementing class as of May 2020 """


	@abstractmethod
	def ingest(self,
		stock_id: StockId,
		datapoints: Sequence[Dict[str, Any]]
	) -> Sequence[T]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsStockT2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 30.04.2020
# Last Modified Date: 08.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Union, Set, Tuple, List, Optional
from ampel.base import abstractmethod
from ampel.type import ChannelId
from ampel.ingest.compile.CompilerBase import CompilerBase


class AbsStockT2Compiler(CompilerBase, abstract=True):

	@abstractmethod
	def compile(self,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
	) -> Dict[Tuple[str, Optional[int]], Set[ChannelId]]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/run/AbsT3UnitRunner.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 04.01.2020
# Last Modified Date: 21.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Sequence, Union, Dict, Any, Optional

from ampel.type import ChannelId
from ampel.base import abstractmethod
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.log.AmpelLogger import AmpelLogger
from ampel.core.AdminUnit import AdminUnit


class AbsT3UnitRunner(AdminUnit, abstract=True):
	"""
	Supply stock views to one or more T3 units.
	"""

	logger: AmpelLogger
	run_id: int
	process_name: str
	channel: Optional[ChannelId] = None

	#: raise exceptions instead of catching and logging
	raise_exc: bool = True
	#: Record the invocation of this event in the journal of each selected stock
	update_journal: bool = True
	#: tag to add to journal records
	extra_journal_tag: Optional[Union[int, str]] = None
	#: contextual information for this run
	run_context: Optional[Dict[str, Any]] = None


	@abstractmethod
	def run(self, data: Sequence[AmpelBuffer]) -> None:
		"""Process a chunk of stocks"""
		...

	@abstractmethod
	def done(self) -> None:
		"""Signal that the run has finished"""
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/type_analysis.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 20.08.2020
# Last Modified Date: 20.08.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any, Iterable, Union, get_origin

def get_subtype(candidate_type : Any, field_type) -> Any:
    """
    Return the first element of `field_type` with which `candidate_type` is
    compatible, or None if no compabible types are found.
    This is essentially a version of issubclass() that handles typing.Union and
    returns the target type instead of just a bool.
    
    :param field_type: a type, sequence of types, or typing.Union
    """
    subfields = None
    if (origin := get_origin(field_type)) is None:
        if isinstance(field_type, Iterable) and not isinstance(field_type, (str, bytes)):
            subfields = field_type 
    elif origin is Union:
        subfields = field_type.__args__
    elif origin is candidate_type:
        # return subscripted generic
        return field_type

    if subfields:
        for subfield in subfields:
            if subtype := get_subtype(candidate_type, subfield):
                return subtype
        else:
            return None
    else:
        try:
            if issubclass(candidate_type, field_type):
                return field_type
        except TypeError:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/run/T3DynamicUnitRunnner.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.01.2020
# Last Modified Date: 21.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Optional, Sequence, Set
from ampel.type import ChannelId
from ampel.model.UnitModel import UnitModel
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.log import AmpelLogger, VERBOSE
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.core import AmpelContext
from ampel.t3.run.AbsT3UnitRunner import AbsT3UnitRunner
from ampel.t3.run.T3UnitRunner import T3UnitRunner


class T3DynamicUnitRunner(AbsT3UnitRunner):
	"""
	Unit runner that for each channel found in the elements loaded by the stages:
	
	- spawns a dedicated :class:`~ampel.t3.run.T3UnitRunner.T3UnitRunner` instance configured to filter and project elements wrt this channel
	- execute the associated T3 units

	Example:
	A general T3 process performs a broad, channel-less query.
	Many stocks are returned, each possibly associated with different channels.
	This unit builds a set of all channels referenced by the results.
	Then, for each channel, the information about the referenced elements are posted into
	a dedicated Slack channel whose name corresponds to the Ampel channel. Suppose,
	for example, that the results contain objects:

	- ``<object 1 with channel {A}>``
	- ``<object 2 with channel {A, B}>``
	- ``<object 3 with channel {B, C, D}>``

	This will cause:
	
	- Object 1 and 2 to be posted to slack channel #A
	- Object 2 and 3  to be posted to slack channel #B
	- Object 3 to be posted to slack channel #C
	- Object 3 to be posted to slack channel #D
	"""

	logger: AmpelLogger
	execute: Sequence[UnitModel]
	white_list: Optional[Set[str]] = None
	black_list: Optional[Set[str]] = None

	remove_empty: bool = True
	unalterable: bool = True
	freeze: bool = False


	def __init__(self, context: AmpelContext, **kwargs) -> None:
		"""
		:param execute: sequence of T3 units to be executed
		:param white_list: perform operations only for channels referenced in this list
		:param black_list: do nothing for channels referenced in this list
		:param channel: (from super class T3UnitRunner): if set, all log entries produced
		by the underlying t3 units will be associated with the provided channel.
		Otherwise, log entries will be associated with the projected channel.
		:param remove_empty: see T3BaseProjector docstring
		:param unalterable: see T3ChannelProjector docstring
		:param freeze: see T3ChannelProjector docstring
		"""

		AmpelBaseModel.__init__(self, **kwargs)
		self.context = context
		self.runners: Dict[ChannelId, T3UnitRunner] = {}

		if self.logger.verbose > 1 and (self.white_list or self.black_list):
			self.warned: Set[ChannelId] = set()

		if self.black_list and self.white_list:
			raise ValueError("Can't have both black and white lists")


	def run(self, data: Sequence[AmpelBuffer]) -> None:

		if not data:
			self.logger.info("No data provided")
			return None

		# step 1: Gather all channels
		# (will raise Error if stock documents were to requested to be loaded)
		channels: Set[ChannelId] = {ell for el in data for ell in el['stock']['channel']}  # type: ignore

		if self.white_list:
			if self.logger.verbose > 1:
				tmp = channels & self.white_list
				for chan in channels - self.white_list:
					if chan not in self.warned:
						self.logger.debug(f"Ignoring channel {chan} (not in white list)")
						self.warned.add(chan)
				channels = tmp
			else:
				channels = channels & self.white_list

		elif self.black_list:
			if self.logger.verbose > 1:
				tmp = channels - self.black_list
				for chan in channels & self.black_list:
					if chan not in self.warned:
						self.logger.debug(f"Ignoring channel {chan} (in black list)")
						self.warned.add(chan)
				channels = tmp
			else:
				channels = channels - self.black_list


		# step 2: spawn T3UnitRunner instances
		for chan in channels:

			if chan not in self.runners:

				if self.logger.verbose:
					self.logger.log(VERBOSE, f"Spawning new T3UnitRunner for channel {chan}")

				self.runners[chan] = T3UnitRunner(
					context = self.context,
					logger = self.logger,
					run_id = self.run_id,
					process_name = self.process_name,
					channel = self.channel if self.channel else chan,
					raise_exc = self.raise_exc,
					update_journal = self.update_journal,
					extra_journal_tag = self.extra_journal_tag,
					run_context = self.run_context,
					directives = [
						{
							"filter": {
								"unit": "T3AmpelBufferFilter",
								"config": {"channel": chan}
							},
							"project": {
								"unit": "T3ChannelProjector",
								"config": {"channel": chan, "remove_empty": self.remove_empty}
							},
							"execute": self.execute
						}
					]
				)


		# step 3: run T3UnitRunner instances
		if self.logger.verbose == 1:
			self.logger.log(VERBOSE, "Running unit runners")

		for chan in channels:

			if self.logger.verbose > 1:
				self.logger.debug(f"Running unit runner for chan {chan}")

			self.runners[chan].run(data)

	def done(self) -> None:
		for runner in self.runners.values():
			runner.done()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/builder/T2UnitIngestionModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.03.2020
# Last Modified Date: 10.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, Union, List
from ampel.model.UnitModel import UnitModel

class T2UnitIngestionModel(UnitModel):
	"""
	UnitModel defines 'resource'
	"""
	ingest: Optional[Dict[str, Any]]
	group: Union[int, List[int]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/test/dummy.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: 11.02.2021
# Last Modified By  : jvs

import time
from collections import defaultdict
from typing import Dict, List, Optional, Sequence, Set, Tuple, Union, Any

from pymongo import UpdateOne

from ampel.abstract.AbsProcessorUnit import AbsProcessorUnit
from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit
from ampel.abstract.AbsTiedStateT2Unit import AbsTiedStateT2Unit
from ampel.abstract.ingest.AbsCompoundIngester import AbsCompoundIngester
from ampel.abstract.ingest.AbsStateT2Compiler import AbsStateT2Compiler
from ampel.abstract.ingest.AbsStateT2Ingester import AbsStateT2Ingester
from ampel.abstract.ingest.AbsT2Ingester import AbsT2Ingester
from ampel.content.DataPoint import DataPoint
from ampel.ingest.CompoundBluePrint import CompoundBluePrint
from ampel.ingest.T1DefaultCombiner import T1DefaultCombiner
from ampel.log.AmpelLogger import AmpelLogger
from ampel.enum.T2SysRunState import T2SysRunState
from ampel.content.T2Document import T2Document
from ampel.type import ChannelId, StockId, T2UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.content.Compound import Compound
from ampel.view.T2DocView import T2DocView
from ampel.model.StateT2Dependency import StateT2Dependency


class Sleepy(AbsProcessorUnit):
    """
    A processor that does nothing (especially not touching the db, which is not
    mocked in subprocesses)
    """

    def run(self):
        time.sleep(1)


class DummyStockT2Unit(AbsStockT2Unit):
    def run(self, stock_doc):
        return {"id": stock_doc["_id"]}


# FIXME: these dummy ingesters are copied from ampel-alerts, as there is no
# default implementation of AbsCompoundIngester or AbsStateT2Ingester
class DummyCompoundIngester(AbsCompoundIngester):
    """simplified PhotoCompoundIngester for testing"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.channels: Set[ChannelId] = set()
        self.engine = T1DefaultCombiner(logger=AmpelLogger.get_logger(console=False))

    def add_channel(self, channel: ChannelId) -> None:
        self.channels.add(channel)

    def ingest(
        self,
        stock_id: StockId,
        datapoints: Sequence[DataPoint],
        chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
    ) -> Optional[CompoundBluePrint]:
        chans = [k for k, v in chan_selection if k in self.channels]

        if not (blue_print := self.engine.combine(stock_id, datapoints, chans)):
            return None

        for eff_comp_id in blue_print.get_effids_for_chans(chans):

            d_addtoset = {
                "channel": {
                    "$each": list(blue_print.get_chans_with_effid(eff_comp_id))
                },
                "run": self.run_id,
            }

            if blue_print.has_flavors(eff_comp_id):
                d_addtoset["flavor"] = {
                    "$each": blue_print.get_compound_flavors(eff_comp_id)
                }

            comp_dict = blue_print.get_eff_compound(eff_comp_id)

            comp_set_on_ins = {
                "_id": eff_comp_id,
                "stock": stock_id,
                "tag": list(blue_print.get_comp_tags(eff_comp_id)),
                "tier": 0,
                "added": time.time(),
                "len": len(comp_dict),
                "body": comp_dict,
            }

            self.updates_buffer.add_t1_update(
                UpdateOne(
                    {"_id": eff_comp_id},
                    {"$setOnInsert": comp_set_on_ins, "$addToSet": d_addtoset},
                    upsert=True,
                )
            )

        return blue_print


class DummyStateT2Compiler(AbsStateT2Compiler):
    def compile(
        self,
        chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
        compound_blueprint: CompoundBluePrint,
    ) -> Dict[
        Tuple[str, Optional[int], Union[bytes, Tuple[bytes, ...]]], Set[ChannelId]
    ]:
        t2s_for_channels = defaultdict(set)
        for chan, ingest_model in self.get_ingest_models(chan_selection):
            t2s_for_channels[(ingest_model.unit_id, ingest_model.config)].add(chan)

        optimized_t2s: Dict[
            Tuple[str, Optional[int], Union[bytes, Tuple[bytes, ...]]], Set[ChannelId]
        ] = {}
        for k, v in t2s_for_channels.items():
            comp_ids = tuple(compound_blueprint.get_effids_for_chans(v))
            if len(comp_ids) == 1:
                optimized_t2s[k + (comp_ids[0],)] = v
            else:
                optimized_t2s[k + (comp_ids,)] = v
        return optimized_t2s


class DummyStateT2Ingester(AbsStateT2Ingester):
    compiler: AbsStateT2Compiler[CompoundBluePrint] = DummyStateT2Compiler()

    def ingest(
        self,
        stock_id: StockId,
        comp_bp: CompoundBluePrint,
        chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
    ) -> None:
        """
        Simplified version of PhotoT2Ingester, with t2 docs linked to exactly
        one compound
        """
        optimized_t2s = self.compiler.compile(chan_selection, comp_bp)
        now = int(time.time())

        # Loop over t2 units to be created
        for (t2_id, run_config, link_id), chans in optimized_t2s.items():

            # Matching search criteria
            match_dict: Dict[str, Any] = {
                "stock": stock_id,
                "unit": t2_id,
                "config": run_config
                # 'link' is added below
            }

            # Attributes set if no previous doc exists
            set_on_insert: T2Document = {
                "stock": stock_id,
                "unit": t2_id,
                "config": run_config,
                "status": T2SysRunState.NEW.value,
            }
            if self.tags:
                set_on_insert["tag"] = self.tags

            jchan, chan_add_to_set = AbsT2Ingester.build_query_parts(chans)
            add_to_set: Dict[str, Any] = {"channel": chan_add_to_set}

            assert isinstance(link_id, bytes)

            match_dict["link"] = {"$elemMatch": {"$eq": link_id}}
            add_to_set["link"] = link_id

            # Update journal
            add_to_set["journal"] = {"tier": self.tier, "dt": now, "channel": jchan}

            # Append update operation to bulk list
            self.updates_buffer.add_t2_update(
                UpdateOne(
                    match_dict,
                    {"$setOnInsert": set_on_insert, "$addToSet": add_to_set},
                    upsert=True,
                )
            )


class DummyPointT2Unit(AbsPointT2Unit):
    def run(self, datapoint):
        return {"thing": datapoint["body"]["thing"]}


class DummyStateT2Unit(AbsStateT2Unit):
    def run(self, compound, datapoints):
        return {"len": len(datapoints)}


class DummyTiedStateT2Unit(AbsTiedStateT2Unit):

    t2_dependency = [StateT2Dependency(unit="DummyStateT2Unit")]
    _unit = "DummyStateT2Unit"

    @classmethod
    def get_tied_unit_names(cls):
        return [cls._unit]

    def run(self,
		compound: Compound,
		datapoints: Sequence[DataPoint],
		t2_records: Sequence[T2DocView]
	) -> T2UnitResult:
        assert t2_records, "dependencies were found"
        assert len(body := t2_records[0].body or []) == 1
        payload = (p[-1] if isinstance((p := t2_records[0].get_payload()), list) else p) # type: ignore[index]
        return {k: v * 2 for k, v in (payload or {}).items()}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/view/plot/T2SVGQuery.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.06.2019
# Last Modified Date: 15.06.2019
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# type: ignore[import]

from ampel.plot.SVGQuery import SVGQuery

class T2SVGQuery(SVGQuery):
	"""
	"""

	def __init__(
		self, col="blend", tran_id=None, plot_tag=None, plot_tags=None,
		t2_class_name=None, t2_run_config=None
	):
		""" """

		super().__init__(
			plot_path='body.output.plots',
			tran_id=tran_id,
			plot_tag=plot_tag,
			plot_tags=plot_tags
		)

		self.col = col

		if t2_class_name:
			self.set_t2_class_name(t2_class_name)

		if t2_run_config:
			self.set_t2_run_config(t2_run_config)


	def set_t2_class_name(self, t2_class_name):
		""" """
		self._query['t2Id'] = t2_class_name


	def set_t2_run_config(self, run_config_name):
		""" """
		self._query['run_config'] = run_config_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/context/T3AddLastRunTime.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.01.2020
# Last Modified Date: 06.08.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, ClassVar
from datetime import datetime, timedelta
from ampel.db.query.events import get_last_run
from ampel.t3.context.AbsT3RunContextAppender import AbsT3RunContextAppender


class T3AddLastRunTime(AbsT3RunContextAppender):

	key: ClassVar[str] = "last_run"

	#: timedelta to search for previous run
	lookup_range: Dict[str, int] = {'days': -7}
	#: timedelta to use if no previous run found
	fallback: Optional[Dict[str, int]] = None  # ex: {'days': -90}

	def update(self, context: Dict[str, Any]) -> None:
		"""Add last run time (UNIX epoch) as "last_run". """
		if self.key in context:
			self.logger.info(f"Field '{self.key}' is already set in context dict")
			return

		last_run = get_last_run(
			self.context.db.get_collection('events'),
			require_success = True,
			process_name = self.process_name,
			gte_time = self.lookup_range
		)

		if last_run is None:
			self.logger.warn(f"Event {self.process_name}: last run time unavailable")
			if self.fallback:
				context[self.key] = (datetime.now() + timedelta(**self.fallback)).timestamp()
				self.logger.warn(f"Fallback last run time: {context[self.key]}")
				return

		context[self.key] = last_run



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/ForwardUnitConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 02.03.2020
# Last Modified Date: 08.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import re, importlib
from typing import List, Union, Dict, Optional, Any, Sequence
from ampel.model.builder.RemoteUnitDefinition import RemoteUnitDefinition
from ampel.config.collector.UnitConfigCollector import UnitConfigCollector
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.AbsForwardConfigCollector import AbsForwardConfigCollector
from ampel.log import VERBOSE


path_el_type: Dict[str, str] = {
	"admin": "AdminUnit",
	"base": "DataUnit",
	"processor": "AbsProcessorUnit",
	"controller": "AbsControllerUnit",
}


class ForwardUnitConfigCollector(AbsForwardConfigCollector):

	def get_path(self,
		arg: Union[Dict[str, Any], str],
		file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> Optional[Sequence[Union[str, int]]]:

		try:

			if isinstance(arg, str):
				class_id = arg
				try:
					mro = self.get_trimmed_mro(arg, UnitConfigCollector.get_class_name(arg))
				except Exception:
					self.error(f"Unit import error: {arg} " + ConfigCollector.distrib_hint(file_name, dist_name))
					return None

			elif isinstance(arg, dict):

				try:
					d = RemoteUnitDefinition(**arg)
					mro = d.base
					class_id = d.class_name
				except Exception:
					# Note: cannot use shorter "return self.error(...)" because of stubborn mypy
					self.error(
						"Unsupported unit definition (dict)" +
						ConfigCollector.distrib_hint(file_name, dist_name)
					)
					return None

			else:
				self.error(
					"Unsupported unit config format" +
					ConfigCollector.distrib_hint(file_name, dist_name)
				)
				return None

			unit_type = None
			for k in path_el_type:
				if path_el_type[k] in mro:
					unit_type = k

			if unit_type is None:
				self.error(f"Could not auto-associate a type with unit {class_id}")
				return None

			# check for DataUnit in mro ?
			if self.verbose:
				self.logger.log(VERBOSE,
					f"-> Routing unit '{class_id}' to unit.{unit_type}"
				)

			return ["unit", unit_type]

		except Exception as e:

			if 'class_id' not in locals():
				class_id = "Unknown"

			self.error(
				f"Error occured while loading unit {class_id} " +
				ConfigCollector.distrib_hint(file_name, dist_name),
				exc_info=e
			)

			return None


	@staticmethod
	def get_class_name(fqn: str) -> str:
		"""
		Returns class name for a given fully qualified name
		Note: we here assume the ampel convention that module name equals class name
		(i.e that we have one class per module)
		"""
		# pylint: disable=anomalous-backslash-in-string
		return re.sub(r".*\.", "", fqn) # noqa


	@staticmethod
	def get_trimmed_mro(mfqn: str, class_name: str) -> List[str]:
		"""
		:param mfqn: fully qualified name of module
		:param class_name: declared class name in the module specified by mfqn
		:returns: the method resolution order (except for the first and last members
		that is of no use for our purpose) of the specified class
		"""
		UnitClass = getattr(
			# import using fully qualified name
			importlib.import_module(mfqn),
			class_name
		)

		return [
			UnitConfigCollector.get_class_name(el.__name__)
			for el in UnitClass.__mro__[1:-1]
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/operator/OneOf.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 22.10.2018
# Last Modified Date: 13.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Generic
from pydantic.generics import GenericModel
from ampel.type import T


class OneOf(GenericModel, Generic[T]):
	#: Select items by logical XOR
	#:
	#: .. warning:: This can't be modelled in Mongo, as it has no logical XOR operation
	one_of: List[T]



from typing import Any, Dict, List, TYPE_CHECKING
import pytest, yaml
import contextlib

from ampel.model.template.AbsLegacyChannelTemplate import AbsLegacyChannelTemplate, T2UnitModel

if TYPE_CHECKING:
    from ampel.log.AmpelLogger import AmpelLogger
    from ampel.config.builder.FirstPassConfig import FirstPassConfig
    from ampel.model.UnitModel import UnitModel

class LegacyChannelTemplate(AbsLegacyChannelTemplate):

    # Mandatory implementation
    def get_processes(
        self, logger: "AmpelLogger", first_pass_config: "FirstPassConfig"
    ) -> List[Dict[str, Any]]:

        ret: List[Dict[str, Any]] = []

        t0_ingester = "DummyAlertContentIngester"
        t1_ingester = "DummyCompoundIngester"
        t2_compute_from_t0 = self.t2_compute
        t2_compute_from_t1: List[T2UnitModel] = []
        ret.insert(
            0,
            self.craft_t0_process(
                first_pass_config,
                controller={},
                stock_ingester="StockIngester",
                t0_ingester=t0_ingester,
                t1_ingester=t1_ingester,
                t1_standalone_ingester=None,
                t2_state_ingester="DummyStateT2Ingester",
                t2_point_ingester="DummyPointT2Ingester",
                t2_stock_ingester="DummyStockT2Ingester",
                t2_compute_from_t0=t2_compute_from_t0,
                t2_compute_from_t1=t2_compute_from_t1,
            ),
        )

        return ret


@pytest.fixture
def first_pass_config(testing_config):
    with open(testing_config, "rb") as f:
        return yaml.safe_load(f)


@pytest.mark.parametrize(
    "t2_compute,target,expected,exception",
    [
        # single, statebound T2
        (
            [{"unit": "DummyStateT2Unit"}],
            ["t0_add", "t1_combine", 0, "t2_compute", "units"],
            {"unit": "DummyStateT2Unit"},
            None
        ),
        # statebound T2 with configured statebound dependency, not requested
        (
            [
                {
                    "unit": "DummyTiedStateT2Unit",
                    "config": {"t2_dependency": [{"unit": "DummyStateT2Unit"}]},
                },
            ],
            ["t0_add", "t1_combine", 0, "t2_compute", "units"],
            {"unit": "DummyStateT2Unit"},
            ValueError,
        ),
        # statebound T2 with implicit (default) statebound dependency
        pytest.param(
            [{"unit": "DummyTiedStateT2Unit"}],
            ["t0_add", "t1_combine", 0, "t2_compute", "units"],
            {"unit": "DummyStateT2Unit"},
            None,
            marks=pytest.mark.xfail(reason="default dependencies aren't automatically resolved")
        ),
        # statebound T2 with statebound dependency, also explicitly configured
        (
            [
                {"unit": "DummyStateT2Unit"},
                {
                    "unit": "DummyTiedStateT2Unit",
                    "config": {"t2_dependency": [{"unit": "DummyStateT2Unit"}]},
                },
            ],
            ["t0_add", "t1_combine", 0, "t2_compute", "units"],
            {"unit": "DummyStateT2Unit"},
            None
        ),
        # statebound T2 with point dependency
        (
            [
                {
                    "unit": "DummyTiedStateT2Unit",
                    "config": {"t2_dependency": [{"unit": "DummyPointT2Unit"}]},
                },
                {"unit": "DummyPointT2Unit"}
            ],
            ["t0_add", "t2_compute", "units"],
            {"unit": "DummyPointT2Unit"},
            None,
        ),
        # statebound T2 with stock dependency
        (
            [
                {
                    "unit": "DummyTiedStateT2Unit",
                    "config": {"t2_dependency": [{"unit": "DummyStockT2Unit"}]},
                },
                {"unit": "DummyStockT2Unit"}
            ],
            ["t2_compute", "units"],
            {"unit": "DummyStockT2Unit"},
            None,
        ),
    ],
)
def test_state_t2_instantiation(t2_compute, target, expected, exception, first_pass_config, monkeypatch):
    """
    Template creates state T2s and checks for missing dependencies
    """
    # _craft_t0_process should raise ValueError if dependencies are missing
    monkeypatch.setattr(
        "ampel.test.dummy.DummyTiedStateT2Unit._unit",
        expected['unit'],
    )
    with (
        pytest.raises(exception)
        if exception
        else contextlib.nullcontext()
    ):
        processes = LegacyChannelTemplate(
            **{
                "channel": "foo",
                "auto_complete": False,
                "t0_filter": {"unit": "DummyFilter"},
                "t2_compute": t2_compute,
            }
        ).get_processes(None, first_pass_config)
    if exception:
        return

    assert processes
    proc = processes[0]
    assert proc["tier"] == 0
    assert (directives := proc["processor"]["config"]["directives"])
    directive = directives[0]

    def get(item, keys):
        while keys:
            item = item[keys.pop(0)]
        return item

    items = get(directive, list(target))
    assert expected in items
    assert (
        len([i for i in items if expected == i]) == 1
    ), "exactly one instance of each unit"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/T2IngestModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 19.11.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any, Dict, List, Union, Optional, Type
from ampel.model.StrictModel import StrictModel
from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit
from ampel.abstract.AbsCustomStateT2Unit import AbsCustomStateT2Unit

class T2IngestModel(StrictModel):

	#: T2 unit to run
	unit: Union[str, Type[AbsStockT2Unit], Type[AbsPointT2Unit], Type[AbsStateT2Unit], Type[AbsCustomStateT2Unit]]
	#: T2 unit configuration (hashed)
	config: Optional[int]
	#: Ingester options
	ingest: Optional[Dict[str, Any]]
	#: Filter result codes that should trigger this T2. If not specified, T2
	#: documents will be created in response to *any* passing alert.
	group: Union[int, List[int]] = []

	@property
	def unit_id(self) -> str:
		if isinstance(self.unit, str):
			return self.unit
		else:
			return self.unit.__name__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/T0AddModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 19.11.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from pydantic import Field
from typing import List, Optional, Type, Union
from ampel.model.UnitModel import UnitModel
from ampel.model.ingest.T1CombineModel import T1CombineModel
from ampel.model.ingest.T2ComputeModel import T2ComputeModel
from ampel.abstract.ingest.AbsIngester import AbsIngester

class T0AddModel(UnitModel):
	# Override 'unit' to enable alias
	#: Add :class:`datapoints <ampel.content.DataPoint.DataPoint>` to the database
	unit: Union[str, Type[AbsIngester]] = Field(..., alias='ingester')
	# config (datapoint ingester config [from UnitModel])
	#: Create :class:`compounds <ampel.content.Compound.Compound>` from ingested :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	t1_combine: Optional[List[T1CombineModel]]
	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>` bound to :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	t2_compute: Optional[T2ComputeModel]



from os import environ
from pathlib import Path

import mongomock
import pytest
from pymongo import InsertOne

from ampel.config.AmpelConfig import AmpelConfig
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.db.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.ingest.PointT2Ingester import PointT2Ingester
from ampel.ingest.StockIngester import StockIngester
from ampel.ingest.StockT2Ingester import StockT2Ingester
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogsBufferDict import LogsBufferDict
from ampel.model.ingest.T2IngestModel import T2IngestModel
from ampel.model.StateT2Dependency import StateT2Dependency
from ampel.test.dummy import DummyCompoundIngester, DummyStateT2Ingester

def pytest_addoption(parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="run docker-based integration tests",
    )

@pytest.fixture(scope="session")
def mongod(pytestconfig):
    if "MONGO_PORT" in environ:
        yield "mongodb://localhost:{}".format(environ["MONGO_PORT"])
        return

    if not pytestconfig.getoption("--integration"):
        raise pytest.skip("integration tests require --integration flag")
    try:
        container = subprocess.check_output(
            "docker", "run", "--rm", "-d", "-P", "mongo:4.4"
        )
    except FileNotFoundError:
        pytest.skip("integration tests require docker")
    try:
        port = json.loads(subprocess.check_output("docker", "inspect", container))[0][
            "NetworkSettings"
        ]["Ports"]["27017/tcp"][0]["HostPort"]
        yield f"mongodb://localhost:{port}"
    finally:
        subprocess.check_call("docker", "stop", "container")


@pytest.fixture
def patch_mongo(monkeypatch):
    monkeypatch.setattr("ampel.db.AmpelDB.MongoClient", mongomock.MongoClient)
    # ignore codec_options in DBContentLoader
    monkeypatch.setattr("mongomock.codec_options.is_supported", lambda *args: None)


@pytest.fixture
def testing_config():
    return Path(__file__).parent / "test-data" / "testing-config.yaml"


@pytest.fixture(scope="session")
def core_config():
    """
    The config distributed with ampel-core
    """
    cb = DistConfigBuilder()
    for ext in "json", "yml", "yaml":
        cb.load_distrib("ampel-core", "conf", ext)
    return cb.build_config(config_validator=None)


@pytest.fixture
def mock_context(patch_mongo, testing_config):
    config = AmpelConfig.load(testing_config)
    return DevAmpelContext.new(config=config, purge_db=True)

@pytest.fixture
def integration_context(mongod, testing_config):
    config = AmpelConfig.load(testing_config)
    return DevAmpelContext.new(config=config, purge_db=True, custom_conf={"resource.mongo": mongod})

# metafixture as suggested in https://github.com/pytest-dev/pytest/issues/349#issuecomment-189370273
@pytest.fixture(params=["mock_context", "integration_context"])
def dev_context(request):
    yield request.getfixturevalue(request.param)

@pytest.fixture
def ampel_logger():
    return AmpelLogger.get_logger()


@pytest.fixture
def ingest_stock(dev_context, ampel_logger):
    run_id = 0
    updates_buffer = DBUpdatesBuffer(dev_context.db, run_id=run_id, logger=ampel_logger)
    logd = LogsBufferDict(
        {
            "logs": [],
            "extra": {},
            "err": False,
        }
    )
    ingester = StockIngester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    )
    ingester.ingest("stockystock", [("TEST_CHANNEL", True)], jextra={"alert": 123})
    ingester.updates_buffer.push_updates()
    assert dev_context.db.get_collection("stock").count_documents({}) == 1


@pytest.fixture
def ingest_stock_t2(dev_context, ampel_logger):
    run_id = 0
    updates_buffer = DBUpdatesBuffer(dev_context.db, run_id=run_id, logger=ampel_logger)
    logd = LogsBufferDict(
        {
            "logs": [],
            "extra": {},
            "err": False,
        }
    )
    ingester = StockT2Ingester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    )
    ingester.add_ingest_models("TEST_CHANNEL", [T2IngestModel(unit="DummyStockT2Unit")])
    ingester.ingest("stockystock", [("TEST_CHANNEL", True)])
    ingester.updates_buffer.push_updates()
    assert dev_context.db.get_collection("t2").count_documents({}) == 1


@pytest.fixture(params=["DummyStateT2Unit", "DummyPointT2Unit", "DummyStockT2Unit"])
def ingest_tied_t2(dev_context, ampel_logger, request, monkeypatch):
    """Create a T2 document with dependencies"""

    tied_config_id = dev_context.add_config_id(
        {"t2_dependency": [{"unit": request.param}]}
    )
    monkeypatch.setattr(
        "ampel.test.dummy.DummyTiedStateT2Unit._unit",
        request.param,
    )

    run_id = 0
    updates_buffer = DBUpdatesBuffer(dev_context.db, run_id=run_id, logger=ampel_logger)
    logd = LogsBufferDict(
        {
            "logs": [],
            "extra": {},
            "err": False,
        }
    )
    filter_results = [("TEST_CHANNEL", True)]

    StockIngester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    ).ingest("stockystock", filter_results, {})

    # mimic a datapoint ingester
    datapoints = [
        {"_id": i, "stock": "stockystock", "body": {"thing": i + 1}} for i in range(3)
    ]
    for dp in datapoints:
        updates_buffer.add_t0_update(InsertOne(dp))

    # create compounds
    comp_ingester = DummyCompoundIngester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    )
    comp_ingester.add_channel(filter_results[0][0])
    blueprint = comp_ingester.ingest("stockystock", datapoints, filter_results)

    # create T2 doc
    # FIXME: should dependent docs be created implicitly?
    state_ingester = DummyStateT2Ingester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    )
    point_ingester = PointT2Ingester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    )
    stock_ingester = StockT2Ingester(
        updates_buffer=updates_buffer,
        logd=logd,
        run_id=run_id,
        context=dev_context,
    )

    # NB: ingestion configured in reverse order so that t2 execution is forced
    # to explicitly resolve dependencies

    state_ingester.add_ingest_models(
        filter_results[0][0],
        [
            T2IngestModel(
                unit="DummyTiedStateT2Unit",
                config=tied_config_id,
            )
        ],
    )

    if "state" in request.param.lower():
        state_ingester.add_ingest_models(
            filter_results[0][0], [T2IngestModel(unit=request.param)]
        )
    elif "point" in request.param.lower():
        point_ingester.add_ingest_models(
            filter_results[0][0], [T2IngestModel(unit=request.param)]
        )
    elif "stock" in request.param.lower():
        stock_ingester.add_ingest_models(
            filter_results[0][0], [T2IngestModel(unit=request.param)]
        )

    state_ingester.ingest("stockystock", blueprint, filter_results)
    point_ingester.ingest("stockystock", datapoints, filter_results)
    stock_ingester.ingest("stockystock", filter_results)

    updates_buffer.push_updates()

    assert (
        dev_context.db.get_collection("t2").count_documents({"unit": request.param})
        == len(datapoints)
        if "point" in request.param.lower()
        else 1
    )
    assert (
        dev_context.db.get_collection("t2").count_documents(
            {"unit": "DummyTiedStateT2Unit"}
        )
        == 1
    )

    return request



import pytest

from ampel.abstract.AbsT3Unit import AbsT3Unit
from ampel.t3.T3Processor import T3Processor


class Mutineer(AbsT3Unit):
    raise_on_add: bool = False
    raise_on_done: bool = False

    def add(self, views):
        if self.raise_on_add:
            raise ValueError

    def done(self):
        if self.raise_on_done:
            raise ValueError


def mutineer_process(config={}):
    return {
        "process_name": "yarrr",
        "chunk_size": 1,
        "directives": [
            {
                "select": {"unit": "T3StockSelector"},
                "load": {
                    "unit": "T3SimpleDataLoader",
                    "config": {"directives": [{"col": "stock"},]},
                },
                "run": {
                    "unit": "T3UnitRunner",
                    "config": {
                        "raise_exc": True,
                        "directives": [
                            {"execute": [{"unit": Mutineer, "config": config}]}
                        ],
                    },
                },
            }
        ],
    }


@pytest.mark.parametrize(
    "config,expect_success",
    [({}, True), ({"raise_on_done": True}, False), ({"raise_on_add": True}, False),],
)
def test_unit_raises_error(dev_context, ingest_stock, config, expect_success):
    """Run is marked failed if units raise an exception"""
    t3 = T3Processor(context=dev_context, raise_exc=False, **mutineer_process(config))
    t3.run()
    assert dev_context.db.get_collection("events").count_documents({}) == 1
    event = dev_context.db.get_collection("events").find_one({})
    assert event["run"] == 1
    assert event["success"] == expect_success



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/T1DefaultCombiner.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.01.2018
# Last Modified Date: 18.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import hashlib, json
from typing import Sequence, List, Union, Generic, TypeVar, Tuple, Optional, Set, Iterable
from ampel.content.DataPoint import DataPoint
from ampel.content.Compound import CompoundElement
from ampel.ingest.CompoundBluePrint import CompoundBluePrint
from ampel.type import StockId, ChannelId, DataPointId
from ampel.abstract.AbsT1Unit import AbsT1Unit
from ampel.log import VERBOSE

T = TypeVar("T", bound=CompoundBluePrint)

class T1DefaultCombiner(Generic[T], AbsT1Unit[T]):
	"""
	Purpose of this class is to generate a "CompoundBluePrint",
	which is a dataclass containing all information required to
	generate a compound document and the t2 documents associated with it.
	Known subclass: PhotoCompoundBuilder from distribution "ampel-photometry".
	Used abbreviations: "eid": effective id, "sid": strict id,
	"chan": channel name, "comp": compound
	"""

	debug: bool = False

	def combine(self,
		stock_id: StockId,
		datapoints: Sequence[DataPoint],
		channel_names: Iterable[ChannelId]
	) -> T:
		"""
		:param datapoints: sequence of dict instances representing photopoint or upperlimis measurements
		!IMPORTANT!: sequence must be timely sorted.
		Sorting is not performed within this method because CompoundBluePrint aims at being instrument independant.
		(For ZTF, the field 'jd' is used: -> sorted(datapoints, key=lambda k: k['jd']))
		"""

		cbp = self.BluePrintClass()
		gen_comp_sub_entry = self.gen_sub_entry # shortcut
		if not isinstance(stock_id, str):
			stock_id_str = str(stock_id)
		else:
			stock_id_str = stock_id

		# Main loop through provided channels
		for chan_name in channel_names:

			# Generate compounds, md5s, and flavors
			#######################################

			# Effective compound contains only non-excluded datapoints
			eff_comp: List[Union[DataPointId, CompoundElement]] = []
			strict_comp: List[Union[DataPointId, CompoundElement]] = []
			eff_hash_payload: List[str] = [stock_id_str]
			strict_hash_payload = stock_id_str
			tags: Set[Union[int, str]] = set()

			# Create compound and compoundId
			for dp in datapoints:

				# Generate compound entry dictionary
				comp_entry, t = gen_comp_sub_entry(dp, chan_name)

				if t:
					tags |= t

				# Append generated dict to effective compound list
				strict_comp.append(comp_entry)

				# Check if policy or exclusion
				# In both case, strict id payload must be appended
				if isinstance(comp_entry, dict):

					# Build str repr of dict. sort_keys is important !
					dict_str = json.dumps(comp_entry, sort_keys=True)

					# Append dict string to strict payload
					strict_hash_payload += dict_str

					# Photopoint or upper limit has a custom policy
					if 'excl' not in comp_entry:

						# For this class, a datapoint defined with policy is just like
						# a different datapoint (meaning a datapoint with a different ID).
						# A datapoint might for example return a different magnitude
						# when associated with a policy.
						# Datapoint with policies will result in a different effective payload
						# and thus a different effective id
						eff_hash_payload.append(dict_str)
						eff_comp.append(comp_entry)

					else:
						if self.debug:
							self.logger.debug("Excluded datapoint", extra=comp_entry) # type: ignore

				# No policy or exclusion
				else:

					# pps or uls ids are long (ZTF-IPAC)
					id_str = str(comp_entry)

					# Append string repr of dict to effective and strict payloads
					eff_comp.append(comp_entry)
					eff_hash_payload.append(id_str)
					strict_hash_payload += id_str

			if not eff_comp:
				self.logger.error(f"Empty compound for channel {chan_name}")
				continue

			# eff_id = effective compound id = md5 hash of effective payload
			# Note: ho is the abbreviation of "hash object"
			eff_id: bytes = hashlib \
				.md5(bytes("".join(eff_hash_payload), "utf-8")) \
				.digest()

			# strict_id = strict compound id = md5 hash of strict payload
			strict_id: bytes = hashlib \
				.md5(bytes(strict_hash_payload, "utf-8")) \
				.digest()

			# Subclasses might need to compute additional values
			self.combine_extra(cbp, chan_name, eff_id, eff_comp, eff_hash_payload)

			# Save results
			##############

			# Save channel name <-> effective comp id association
			if eff_id in cbp.d_eid_chnames:
				cbp.d_eid_chnames[eff_id].add(chan_name)
				cbp.d_eid_comptags[eff_id] |= tags
			else:
				cbp.d_eid_chnames[eff_id] = {chan_name}
				cbp.d_eid_comptags[eff_id] = tags

			# Save effective compound
			cbp.d_eid_comp[eff_id] = eff_comp

			# If effective id does not equal strict id, a compound flavor entry
			# should to be created in the compound document
			if eff_id != strict_id:
				if self.logger:
					self.logger.log(VERBOSE,
						None, extra = {
							'channel': chan_name,
							'eff': eff_id,
							'strict': strict_id
						}
					)

				# Add tupple (chan_name, strict id) to internal dict using eff_id as key
				if eff_id not in cbp.d_eid_tuple_chan_sid:
					cbp.d_eid_tuple_chan_sid[eff_id] = {(chan_name, strict_id)}
				else:
					cbp.d_eid_tuple_chan_sid[eff_id].add((chan_name, strict_id))

				# Save strict compound using strict id as key
				cbp.d_sid_compdiff[strict_id] = strict_comp

		# Return instance of CompoundBluePrint
		return cbp if cbp.d_eid_comp else None


	def gen_sub_entry(self,
		dp: DataPoint, channel_name: ChannelId
	) -> Tuple[Union[DataPointId, CompoundElement], Optional[Set[Union[int, str]]]]:
		"""
		Method can be overriden by subclasses.
		Known overriding class: ZiCompoundBuilder (distrib ampel-ZTF)
		"""

		# Channel specific exclusion. dp["excl"] could look like this: ["HU_SN", "HU_GRB"]
		if "excl" in dp and channel_name in dp['excl']: # type: ignore
				return {'id': dp['_id'], 'excl': 'Manual'}, {'HAS_EXCLUDED_PPS', 'MANUAL_EXCLUSION'}

		return dp['_id'], None


	def combine_extra(self,
		blue_print: T, chan_name: ChannelId, eff_id: bytes,
		eff_comp: List[Union[DataPointId, CompoundElement]],
		eff_hash_payload: List[str]
	) -> None:
		"""
		Method can be overriden by subclasses.
		Known overriding class: T1PhotoCompoundBuilder (distrib ampel-photometry)
		"""
		pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/plot/SVGLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.06.2019
# Last Modified Date: 03.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# type: ignore[import]

from ampel.db.AmpelDB import AmpelDB
from ampel.plot.SVGCollection import SVGCollection
from ampel.plot.SVGQuery import SVGQuery
from ampel.plot.T2SVGQuery import T2SVGQuery
from collections import defaultdict

class SVGLoader:

	def __init__(self, data_query=None, t2_query=None):
		self._data_query = data_query
		self._t2_query = t2_query
		self._plots = defaultdict(SVGCollection)


	@staticmethod
	def load_all(
		tran_id=None, plot_tag=None, plot_tags=None,
		t2_class_name=None, t2_run_config=None
	):

		data_query = SVGQuery(
			tran_id = tran_id,
			plot_tag = plot_tag,
			plot_tags = plot_tags
		)

		t2_query = T2SVGQuery(
			tran_id = tran_id,
			plot_tag = plot_tag,
			plot_tags = plot_tags,
			t2_class_name = t2_class_name,
			t2_run_config = t2_run_config
		)

		sl = SVGLoader(
			data_query = data_query,
			t2_query = t2_query
		)

		sl.load_plots()

		return sl


	def load_plots(self):

		if self._data_query is not None:

			for el in AmpelDB.get_collection("t0").find(self._data_query._query):
				self._load_plots(
					el['stockId'],
					self._data_query, el['plots']
				)

		if self._t2_query is not None:

			for el in AmpelDB.get_collection("t2").find(self._t2_query._query):
				if 'body' not in el or not el['body']:
					continue
				if 'output' in el['body'][-1] and 'plots' in el['body'][-1]['output']:
					self._load_plots(
						el['stockId'],
						self._t2_query,
						el['body'][-1]['output']['plots']
					)


	def _load_plots(self, tran_id, query, plots):

		for p in plots:

			if query.tag:
				if query.tag not in p['tag']:
					continue

			if query.tags:
				if not all(x in p['tag'] for x in query.tags):
					continue

			self._plots[tran_id].add_raw_db_dict(p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/plot/SVGPlot.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.06.2019
# Last Modified Date: 09.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# type: ignore[import]

from types import List, Optional
from ampel.type import StockId
from ampel.plot.SVGUtils import SVGUtils

class SVGPlot:

	#display_div = '<div style="display: inline-flex; justify-content: center;">'

	def __init__(
		self, content, scale: float = 1.0, title: Optional[str] = None,
		tags: Optional[List[str]] = None, stock_id: Optional[StockId] = None,
		title_left_padding: int = 0, center: bool = True
	):

		self._title = title
		self._content = content
		self._scale = scale
		self._tags = tags
		self._title_left_padding = title_left_padding
		self._center = center
		self._stock_id = stock_id


	def rescale(self, scale: float = 1.0) -> None:

		if self._scale == scale:
			return

		self._content = SVGUtils.rescale(self._content, scale)
		self._scale = scale


	def has_tag(self, tag: str) -> bool:
		return tag in self._tags


	def has_tags(self, tags: List[str]) -> bool:
		return all(el in self._tags for el in tags)


	def _repr_html_(
		self, scale: Optional[float] = None, show_title: bool = True,
		title_prefix: Optional[str] = None, show_tags: bool = False,
		padding_bottom: int = 0, png_convert: bool = False
	) -> str:
		"""
		:param scale: if None, native scaling is used
		"""

		html = "<center>" if self._center else ""

		html += '<div style="padding-bottom: %ipx">' % padding_bottom

		if show_title:
			html += '<h3 style="padding-left:%ipx">%s %s</h3>' % (
				self._title_left_padding,
				"" if title_prefix is None else title_prefix,
				self._title
			)

		if show_tags:
			html += '<h3>' + str(self._tags) + '</h3>'

		#html += SVGPlot.display_div
		html += "<div>"

		if scale is not None and isinstance(scale, (float, int)):
			if png_convert:
				html += SVGUtils.to_png(
					SVGUtils.rescale(self._content, scale),
					dpi=png_convert
				)
			else:
				html += SVGUtils.rescale(self._content, scale)
		else:
			html += SVGUtils.to_png(self._content, dpi=png_convert) if png_convert else self._content

		return html + '</div></div></center>' if self._center else '</div></div>'


	def show_html(self, **kwargs):
		"""
		:param **kwargs: see _repr_html_ arguments for details
		"""
		from IPython.display import HTML
		return HTML(
			self._repr_html_(**kwargs)
		)
